# -*- coding: utf-8 -*-
"""
Unit tests for the signal_processing module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function

import unittest

import neo
import numpy as np
import quantities as pq
import scipy.signal as spsig
import scipy.stats
from numpy.ma.testutils import assert_array_equal, assert_allclose
from numpy.testing import assert_array_almost_equal

import elephant.signal_processing


class PairwiseCrossCorrelationTest(unittest.TestCase):
    # Set parameters
    sampling_period = 0.02 * pq.s
    sampling_rate = 1. / sampling_period
    n_samples = 2018
    times = np.arange(n_samples) * sampling_period
    freq = 1. * pq.Hz

    def test_cross_correlation_freqs(self):
        """
        Sine vs cosine for different frequencies
        Note, that accuracy depends on N and min(f).
        E.g., f=0.1 and N=2018 only has an accuracy on the order decimal=1
        """
        freq_arr = np.linspace(0.5, 15, 8) * pq.Hz
        signal = np.zeros((self.n_samples, 3))
        for freq in freq_arr:
            signal[:, 0] = np.sin(2. * np.pi * freq * self.times)
            signal[:, 1] = np.cos(2. * np.pi * freq * self.times)
            signal[:, 2] = np.cos(2. * np.pi * freq * self.times + 0.2)
            # Convert signal to neo.AnalogSignal
            signal_neo = neo.AnalogSignal(signal, units='mV',
                                          t_start=0. * pq.ms,
                                          sampling_rate=self.sampling_rate,
                                          dtype=float)
            rho = elephant.signal_processing.cross_correlation_function(
                signal_neo, [[0, 1], [0, 2]])
            # Cross-correlation of sine and cosine should be sine
            assert_array_almost_equal(
                rho.magnitude[:, 0], np.sin(2. * np.pi * freq * rho.times),
                decimal=2)
            self.assertEqual(rho.shape, (signal.shape[0], 2))  # 2 pairs

    def test_cross_correlation_nlags(self):
        """
        Sine vs cosine for specific nlags
        """
        nlags = 30
        signal = np.zeros((self.n_samples, 2))
        signal[:, 0] = 0.2 * np.sin(2. * np.pi * self.freq * self.times)
        signal[:, 1] = 5.3 * np.cos(2. * np.pi * self.freq * self.times)
        # Convert signal to neo.AnalogSignal
        signal = neo.AnalogSignal(signal, units='mV', t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate,
                                  dtype=float)
        rho = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1], n_lags=nlags)
        # Test if vector of lags tau has correct length
        assert len(rho.times) == 2 * int(nlags) + 1

    def test_cross_correlation_phi(self):
        """
        Sine with phase shift phi vs cosine
        """
        phi = np.pi / 6.
        signal = np.zeros((self.n_samples, 2))
        signal[:, 0] = 0.2 * np.sin(2. * np.pi * self.freq * self.times + phi)
        signal[:, 1] = 5.3 * np.cos(2. * np.pi * self.freq * self.times)
        # Convert signal to neo.AnalogSignal
        signal = neo.AnalogSignal(signal, units='mV', t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate,
                                  dtype=float)
        rho = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1])
        # Cross-correlation of sine and cosine should be sine + phi
        assert_array_almost_equal(rho.magnitude[:, 0], np.sin(
            2. * np.pi * self.freq * rho.times + phi), decimal=2)

    def test_cross_correlation_envelope(self):
        """
        Envelope of sine vs cosine
        """
        # Sine with phase shift phi vs cosine for different frequencies
        nlags = 800  # nlags need to be smaller than N/2 b/c border effects
        signal = np.zeros((self.n_samples, 2))
        signal[:, 0] = 0.2 * np.sin(2. * np.pi * self.freq * self.times)
        signal[:, 1] = 5.3 * np.cos(2. * np.pi * self.freq * self.times)
        # Convert signal to neo.AnalogSignal
        signal = neo.AnalogSignal(signal, units='mV', t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate,
                                  dtype=float)
        envelope = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1], n_lags=nlags, hilbert_envelope=True)
        # Envelope should be one for sinusoidal function
        assert_array_almost_equal(envelope, np.ones_like(envelope), decimal=2)

    def test_cross_correlation_biased(self):
        signal = np.c_[np.sin(2. * np.pi * self.freq * self.times),
                       np.cos(2. * np.pi * self.freq * self.times)] * pq.mV
        signal = neo.AnalogSignal(signal, t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate)
        raw = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1], scaleopt='none'
        )
        biased = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1], scaleopt='biased'
        )
        assert_array_almost_equal(biased, raw / biased.shape[0])

    def test_cross_correlation_coeff(self):
        signal = np.c_[np.sin(2. * np.pi * self.freq * self.times),
                       np.cos(2. * np.pi * self.freq * self.times)] * pq.mV
        signal = neo.AnalogSignal(signal, t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate)
        normalized = elephant.signal_processing.cross_correlation_function(
            signal, [0, 1], scaleopt='coeff'
        )
        sig1, sig2 = signal.magnitude.T
        target_numpy = np.correlate(sig1, sig2, mode="same")
        target_numpy /= np.sqrt((sig1 ** 2).sum() * (sig2 ** 2).sum())
        target_numpy = np.expand_dims(target_numpy, axis=1)
        assert_array_almost_equal(normalized.magnitude,
                                  target_numpy,
                                  decimal=3)

    def test_cross_correlation_coeff_autocorr(self):
        # Numpy/Matlab equivalent
        signal = np.sin(2. * np.pi * self.freq * self.times)
        signal = signal[:, np.newaxis] * pq.mV
        signal = neo.AnalogSignal(signal, t_start=0. * pq.ms,
                                  sampling_rate=self.sampling_rate)
        normalized = elephant.signal_processing.cross_correlation_function(
            signal, [0, 0], scaleopt='coeff'
        )
        # auto-correlation at zero lag should equal 1
        self.assertAlmostEqual(normalized[normalized.shape[0] // 2], 1)


class ZscoreTestCase(unittest.TestCase):

    def setUp(self):
        self.test_seq1 = [1, 28, 4, 47, 5, 16, 2, 5, 21, 12,
                          4, 12, 59, 2, 4, 18, 33, 25, 2, 34,
                          4, 1, 1, 14, 8, 1, 10, 1, 8, 20,
                          5, 1, 6, 5, 12, 2, 8, 8, 2, 8,
                          2, 10, 2, 1, 1, 2, 15, 3, 20, 6,
                          11, 6, 18, 2, 5, 17, 4, 3, 13, 6,
                          1, 18, 1, 16, 12, 2, 52, 2, 5, 7,
                          6, 25, 6, 5, 3, 15, 4, 3, 16, 3,
                          6, 5, 24, 21, 3, 3, 4, 8, 4, 11,
                          5, 7, 5, 6, 8, 11, 33, 10, 7, 4]
        self.test_seq2 = [6, 3, 0, 0, 18, 4, 14, 98, 3, 56,
                          7, 4, 6, 9, 11, 16, 13, 3, 2, 15,
                          24, 1, 0, 7, 4, 4, 9, 24, 12, 11,
                          9, 7, 9, 8, 5, 2, 7, 12, 15, 17,
                          3, 7, 2, 1, 0, 17, 2, 6, 3, 32,
                          22, 19, 11, 8, 5, 4, 3, 2, 7, 21,
                          24, 2, 5, 10, 11, 14, 6, 8, 4, 12,
                          6, 5, 2, 22, 25, 19, 16, 22, 13, 2,
                          19, 20, 17, 19, 2, 4, 1, 3, 5, 23,
                          20, 15, 4, 7, 10, 14, 15, 15, 20, 1]

    def test_zscore_single_dup(self):
        """
        Test z-score on a single AnalogSignal, asking to return a
        duplicate.
        """
        signal = neo.AnalogSignal(
            self.test_seq1, units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)

        m = np.mean(self.test_seq1)
        s = np.std(self.test_seq1)
        target = (self.test_seq1 - m) / s
        assert_array_equal(target, scipy.stats.zscore(self.test_seq1))

        result = elephant.signal_processing.zscore(signal, inplace=False)
        assert_array_almost_equal(
            result.magnitude, target.reshape(-1, 1), decimal=9)

        self.assertEqual(result.units, pq.Quantity(1. * pq.dimensionless))

        # Assert original signal is untouched
        self.assertEqual(signal[0].magnitude, self.test_seq1[0])

        # Assert original and returned objects are different
        self.assertIsNot(result, signal)

    def test_zscore_single_inplace(self):
        """
        Test z-score on a single AnalogSignal, asking for an inplace
        operation.
        """
        signal = neo.AnalogSignal(
            self.test_seq1, units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)

        m = np.mean(self.test_seq1)
        s = np.std(self.test_seq1)
        target = (self.test_seq1 - m) / s

        result = elephant.signal_processing.zscore(signal, inplace=True)

        assert_array_almost_equal(
            result.magnitude, target.reshape(-1, 1), decimal=9)

        self.assertEqual(result.units, pq.Quantity(1. * pq.dimensionless))

        # Assert original signal is overwritten
        self.assertEqual(signal[0].magnitude, target[0])

        # Assert original and returned objects are the same
        self.assertIs(result, signal)

    def test_zscore_single_multidim_dup(self):
        """
        Test z-score on a single AnalogSignal with multiple dimensions, asking
        to return a duplicate.
        """
        signal = neo.AnalogSignal(
            np.transpose(
                np.vstack([self.test_seq1, self.test_seq2])), units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)

        m = np.mean(signal.magnitude, axis=0, keepdims=True)
        s = np.std(signal.magnitude, axis=0, keepdims=True)
        target = (signal.magnitude - m) / s

        result = elephant.signal_processing.zscore(signal, inplace=False)
        assert_array_almost_equal(result.magnitude, target, decimal=9)

        # Assert original signal is untouched
        self.assertEqual(signal[0, 0].magnitude, self.test_seq1[0])

        # Assert original and returned objects are different
        self.assertIsNot(result, signal)

    def test_zscore_array_annotations(self):
        signal = neo.AnalogSignal(
            self.test_seq1, units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz,
            array_annotations=dict(valid=True, my_list=[0]))
        zscored = elephant.signal_processing.zscore(signal, inplace=False)
        self.assertDictEqual(signal.array_annotations,
                             zscored.array_annotations)

    def test_zscore_single_multidim_inplace(self):
        """
        Test z-score on a single AnalogSignal with multiple dimensions, asking
        for an inplace operation.
        """
        signal = neo.AnalogSignal(
            np.vstack([self.test_seq1, self.test_seq2]), units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)

        m = np.mean(signal.magnitude, axis=0, keepdims=True)
        s = np.std(signal.magnitude, axis=0, keepdims=True)
        ground_truth = np.divide(signal.magnitude - m, s,
                                 out=np.zeros_like(signal.magnitude),
                                 where=s != 0)
        result = elephant.signal_processing.zscore(signal, inplace=True)

        assert_array_almost_equal(result.magnitude, ground_truth, decimal=8)

        # Assert original signal is overwritten
        self.assertAlmostEqual(signal[0, 0].magnitude, ground_truth[0, 0])

        # Assert original and returned objects are the same
        self.assertIs(result, signal)

    def test_zscore_single_dup_int(self):
        """
        Test if the z-score is correctly calculated even if the input is an
        AnalogSignal of type int, asking for a duplicate (duplicate should
        be of type float).
        """
        signal = neo.AnalogSignal(
            self.test_seq1, units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=int)

        m = np.mean(self.test_seq1)
        s = np.std(self.test_seq1)
        target = (self.test_seq1 - m) / s

        result = elephant.signal_processing.zscore(signal, inplace=False)
        assert_array_almost_equal(result.magnitude, target.reshape(-1, 1),
                                  decimal=9)

        # Assert original signal is untouched
        self.assertEqual(signal.magnitude[0], self.test_seq1[0])

        # Assert original and returned objects are different
        self.assertIsNot(result, signal)

    def test_zscore_single_inplace_int(self):
        """
        Test if the z-score operation fails if the input is an
        AnalogSignal of type int, when asking for an inplace operation.
        """

        signal = neo.AnalogSignal(
            self.test_seq1, units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=int)

        with self.assertRaises(ValueError):
            elephant.signal_processing.zscore(signal, inplace=True)

    def test_zscore_list_dup(self):
        """
        Test zscore on a list of AnalogSignal objects, asking to return a
        duplicate.
        """
        signal1 = neo.AnalogSignal(
            np.transpose(np.vstack([self.test_seq1, self.test_seq1])),
            units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)
        signal2 = neo.AnalogSignal(
            np.transpose(np.vstack([self.test_seq1, self.test_seq2])),
            units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)
        signal_list = [signal1, signal2]

        m = np.mean(np.hstack([self.test_seq1, self.test_seq1]))
        s = np.std(np.hstack([self.test_seq1, self.test_seq1]))
        target11 = (self.test_seq1 - m) / s
        target21 = (self.test_seq1 - m) / s
        m = np.mean(np.hstack([self.test_seq1, self.test_seq2]))
        s = np.std(np.hstack([self.test_seq1, self.test_seq2]))
        target12 = (self.test_seq1 - m) / s
        target22 = (self.test_seq2 - m) / s

        # Call elephant function
        result = elephant.signal_processing.zscore(signal_list, inplace=False)

        assert_array_almost_equal(
            result[0].magnitude,
            np.transpose(np.vstack([target11, target12])), decimal=9)
        assert_array_almost_equal(
            result[1].magnitude,
            np.transpose(np.vstack([target21, target22])), decimal=9)

        # Assert original signal is untouched
        self.assertEqual(signal1.magnitude[0, 0], self.test_seq1[0])
        self.assertEqual(signal2.magnitude[0, 1], self.test_seq2[0])

        # Assert original and returned objects are different
        self.assertIsNot(result[0], signal_list[0])
        self.assertIsNot(result[1], signal_list[1])

    def test_zscore_list_inplace(self):
        """
        Test zscore on a list of AnalogSignal objects, asking for an
        inplace operation.
        """
        signal1 = neo.AnalogSignal(
            np.transpose(np.vstack([self.test_seq1, self.test_seq1])),
            units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)
        signal2 = neo.AnalogSignal(
            np.transpose(np.vstack([self.test_seq1, self.test_seq2])),
            units='mV',
            t_start=0. * pq.ms, sampling_rate=1000. * pq.Hz, dtype=float)
        signal_list = [signal1, signal2]

        m = np.mean(np.hstack([self.test_seq1, self.test_seq1]))
        s = np.std(np.hstack([self.test_seq1, self.test_seq1]))
        target11 = (self.test_seq1 - m) / s
        target21 = (self.test_seq1 - m) / s
        m = np.mean(np.hstack([self.test_seq1, self.test_seq2]))
        s = np.std(np.hstack([self.test_seq1, self.test_seq2]))
        target12 = (self.test_seq1 - m) / s
        target22 = (self.test_seq2 - m) / s

        # Call elephant function
        result = elephant.signal_processing.zscore(signal_list, inplace=True)

        assert_array_almost_equal(
            result[0].magnitude,
            np.transpose(np.vstack([target11, target12])), decimal=9)
        assert_array_almost_equal(
            result[1].magnitude,
            np.transpose(np.vstack([target21, target22])), decimal=9)

        # Assert original signal is overwritten
        self.assertEqual(signal1[0, 0].magnitude, target11[0])
        self.assertEqual(signal2[0, 0].magnitude, target21[0])

        # Assert original and returned objects are the same
        self.assertIs(result[0], signal_list[0])
        self.assertIs(result[1], signal_list[1])

    def test_z_score_wrong_input(self):
        # wrong type
        self.assertRaises(TypeError, elephant.signal_processing.zscore,
                          signal=[1, 2] * pq.uV)
        # units mismatch
        asig1 = neo.AnalogSignal([0, 1], units=pq.uV, sampling_rate=1 * pq.ms)
        asig2 = neo.AnalogSignal([0, 1], units=pq.V, sampling_rate=1 * pq.ms)
        self.assertRaises(ValueError, elephant.signal_processing.zscore,
                          signal=[asig1, asig2])

    def test_z_score_np_float32_64(self):
        """
        Regression test: Inplace operations for z_score failed when using
        np.float32 or np.float64 types.
        See Issue #591.
        https://github.com/NeuralEnsemble/elephant/issues/591
        """
        test_types = (np.float32, np.float64)
        for test_type in test_types:
            with self.subTest(test_type):
                signal = neo.AnalogSignal(self.test_seq1, units='mV',
                                          t_start=0. * pq.ms,
                                          sampling_rate=1000. * pq.Hz,
                                          dtype=test_type)
                # This should not raise a ValueError
                elephant.signal_processing.zscore(signal, inplace=True)


class ButterTestCase(unittest.TestCase):

    def test_butter_filter_type(self):
        """
        Test if correct type of filtering is performed according to how cut-off
        frequencies are given
        """
        # generate white noise AnalogSignal
        noise = neo.AnalogSignal(
            np.random.normal(size=5000),
            sampling_rate=1000 * pq.Hz, units='mV')

        # test high-pass filtering: power at the lowest frequency
        # should be almost zero
        # Note: the default detrend function of scipy.signal.welch() seems to
        # cause artificial finite power at the lowest frequencies. Here I avoid
        # this by using an identity function for detrending
        filtered_noise = elephant.signal_processing.butter(
            noise, 250.0 * pq.Hz, None)
        _, psd = spsig.welch(filtered_noise.T, nperseg=1024, fs=1000.0,
                             detrend=lambda x: x)
        self.assertAlmostEqual(psd[0, 0], 0)

        # test low-pass filtering: power at the highest frequency
        # should be almost zero
        filtered_noise = elephant.signal_processing.butter(
            noise, None, 250.0 * pq.Hz)
        _, psd = spsig.welch(filtered_noise.T, nperseg=1024, fs=1000.0)
        self.assertAlmostEqual(psd[0, -1], 0)

        # test band-pass filtering: power at the lowest and highest frequencies
        # should be almost zero
        filtered_noise = elephant.signal_processing.butter(
            noise, 200.0 * pq.Hz, 300.0 * pq.Hz)
        _, psd = spsig.welch(filtered_noise.T, nperseg=1024, fs=1000.0,
                             detrend=lambda x: x)
        self.assertAlmostEqual(psd[0, 0], 0)
        self.assertAlmostEqual(psd[0, -1], 0)

        # test band-stop filtering: power at the intermediate frequency
        # should be almost zero
        filtered_noise = elephant.signal_processing.butter(
            noise, 400.0 * pq.Hz, 100.0 * pq.Hz)
        _, psd = spsig.welch(filtered_noise.T, nperseg=1024, fs=1000.0)
        self.assertAlmostEqual(psd[0, 256], 0)

    def test_butter_filter_function(self):
        """
        `elephant.signal_processing.butter` return values test for all
        available filters (result has to be almost equal):
            * lfilter
            * filtfilt
            * sosfiltfilt
        """
        # generate white noise AnalogSignal
        noise = neo.AnalogSignal(
            np.random.normal(size=5000),
            sampling_rate=1000 * pq.Hz, units='mV',
            array_annotations=dict(valid=True, my_list=[0]))

        kwds = {'signal': noise, 'highpass_frequency': 250.0 * pq.Hz,
                'lowpass_frequency': None, 'filter_function': 'filtfilt'}
        filtered_noise = elephant.signal_processing.butter(**kwds)
        _, psd_filtfilt = spsig.welch(
            filtered_noise.T, nperseg=1024, fs=1000.0, detrend=lambda x: x)

        kwds['filter_function'] = 'lfilter'
        filtered_noise = elephant.signal_processing.butter(**kwds)
        _, psd_lfilter = spsig.welch(
            filtered_noise.T, nperseg=1024, fs=1000.0, detrend=lambda x: x)

        kwds['filter_function'] = 'sosfiltfilt'
        filtered_noise = elephant.signal_processing.butter(**kwds)
        _, psd_sosfiltfilt = spsig.welch(
            filtered_noise.T, nperseg=1024, fs=1000.0, detrend=lambda x: x)

        self.assertAlmostEqual(psd_filtfilt[0, 0], psd_lfilter[0, 0])
        self.assertAlmostEqual(psd_filtfilt[0, 0], psd_sosfiltfilt[0, 0])

        # Test if array_annotations are preserved
        self.assertDictEqual(noise.array_annotations,
                             filtered_noise.array_annotations)

    def test_butter_invalid_filter_function(self):
        # generate a dummy AnalogSignal
        anasig_dummy = neo.AnalogSignal(
            np.zeros(5000), sampling_rate=1000 * pq.Hz, units='mV')
        # test exception upon invalid filtfunc string
        kwds = {'signal': anasig_dummy, 'highpass_frequency': 250.0 * pq.Hz,
                'filter_function': 'invalid_filter'}
        self.assertRaises(
            ValueError, elephant.signal_processing.butter, **kwds)

    def test_butter_missing_cutoff_freqs(self):
        # generate a dummy AnalogSignal
        anasig_dummy = neo.AnalogSignal(
            np.zeros(5000), sampling_rate=1000 * pq.Hz, units='mV')
        # test a case where no cut-off frequencies are given
        kwds = {'signal': anasig_dummy, 'highpass_frequency': None,
                'lowpass_frequency': None}
        self.assertRaises(
            ValueError, elephant.signal_processing.butter, **kwds)

    def test_butter_input_types(self):
        # generate white noise data of different types
        noise_np = np.random.normal(size=5000)
        noise_pq = noise_np * pq.mV
        noise = neo.AnalogSignal(noise_pq, sampling_rate=1000.0 * pq.Hz)

        # check input as NumPy ndarray
        filtered_noise_np = elephant.signal_processing.butter(
            noise_np, 400.0, 100.0, sampling_frequency=1000.0)
        self.assertTrue(isinstance(filtered_noise_np, np.ndarray))
        self.assertFalse(isinstance(filtered_noise_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(filtered_noise_np, neo.AnalogSignal))
        self.assertEqual(filtered_noise_np.shape, noise_np.shape)

        # check input as Quantity array
        filtered_noise_pq = elephant.signal_processing.butter(
            noise_pq, 400.0 * pq.Hz, 100.0 * pq.Hz, sampling_frequency=1000.0)
        self.assertTrue(isinstance(filtered_noise_pq, pq.quantity.Quantity))
        self.assertFalse(isinstance(filtered_noise_pq, neo.AnalogSignal))
        self.assertEqual(filtered_noise_pq.shape, noise_pq.shape)

        # check input as neo AnalogSignal
        filtered_noise = elephant.signal_processing.butter(noise,
                                                           400.0 * pq.Hz,
                                                           100.0 * pq.Hz)
        self.assertTrue(isinstance(filtered_noise, neo.AnalogSignal))
        self.assertEqual(filtered_noise.shape, noise.shape)

        # check if the results from different input types are identical
        self.assertTrue(np.all(
            filtered_noise_pq.magnitude == filtered_noise_np))
        self.assertTrue(np.all(
            filtered_noise.magnitude[:, 0] == filtered_noise_np))

    def test_butter_axis(self):
        noise = np.random.normal(size=(4, 5000))
        filtered_noise = elephant.signal_processing.butter(
            noise, 250.0, sampling_frequency=1000.0)
        filtered_noise_transposed = elephant.signal_processing.butter(
            noise.T, 250.0, sampling_frequency=1000.0, axis=0)
        self.assertTrue(np.all(filtered_noise == filtered_noise_transposed.T))

    def test_butter_multidim_input(self):
        noise_pq = np.random.normal(size=(4, 5000)) * pq.mV
        noise_neo = neo.AnalogSignal(
            noise_pq.T, sampling_rate=1000.0 * pq.Hz)
        noise_neo1d = neo.AnalogSignal(
            noise_pq[0], sampling_rate=1000.0 * pq.Hz)
        filtered_noise_pq = elephant.signal_processing.butter(
            noise_pq, 250.0, sampling_frequency=1000.0)
        filtered_noise_neo = elephant.signal_processing.butter(
            noise_neo, 250.0)
        filtered_noise_neo1d = elephant.signal_processing.butter(
            noise_neo1d, 250.0)
        self.assertTrue(np.all(
            filtered_noise_pq.magnitude == filtered_noise_neo.T.magnitude))
        self.assertTrue(np.all(
            filtered_noise_neo1d.magnitude[:, 0] ==
            filtered_noise_neo.magnitude[:, 0]))


class HilbertTestCase(unittest.TestCase):

    def setUp(self):
        # Generate test data of a harmonic function over a long time
        time = np.arange(0, 1000, 0.1) * pq.ms
        freq = 10 * pq.Hz

        self.amplitude = np.array([
            np.linspace(1, 10, len(time)),
            np.linspace(1, 10, len(time)),
            np.ones((len(time))),
            np.ones((len(time))) * 10.]).T
        self.phase = np.array([
            (time * freq).simplified.magnitude * 2. * np.pi,
            (time * freq).simplified.magnitude * 2. * np.pi + np.pi / 2,
            (time * freq).simplified.magnitude * 2. * np.pi + np.pi,
            (time * freq).simplified.magnitude * 2. * 2. * np.pi]).T

        self.phase = np.mod(self.phase + np.pi, 2. * np.pi) - np.pi

        # rising amplitude cosine, random ampl. sine, flat inverse cosine,
        # flat cosine at double frequency
        sigs = np.vstack([
            self.amplitude[:, 0] * np.cos(self.phase[:, 0]),
            self.amplitude[:, 1] * np.cos(self.phase[:, 1]),
            self.amplitude[:, 2] * np.cos(self.phase[:, 2]),
            self.amplitude[:, 3] * np.cos(self.phase[:, 3])])

        array_annotations = dict(my_list=np.arange(sigs.shape[0]))
        self.long_signals = neo.AnalogSignal(
            sigs.T, units='mV',
            t_start=0. * pq.ms,
            sampling_rate=(len(time) / (time[-1] - time[0])).rescale(pq.Hz),
            dtype=float,
            array_annotations=array_annotations)

        # Generate test data covering a single oscillation cycle in 1s only
        phases = np.arange(0, 2 * np.pi, np.pi / 256)
        sigs = np.vstack([
            np.sin(phases),
            np.cos(phases),
            np.sin(2 * phases),
            np.cos(2 * phases)])

        self.one_period = neo.AnalogSignal(
            sigs.T, units=pq.mV,
            sampling_rate=len(phases) * pq.Hz)

    def test_hilbert_pad_type_error(self):
        """
        Tests if incorrect pad_type raises ValueError.
        """
        padding = 'wrong_type'

        self.assertRaises(
            ValueError, elephant.signal_processing.hilbert,
            self.long_signals, padding=padding)

    def test_hilbert_output_shape(self):
        """
        Tests if the length of the output is identical to the original signal,
        and the dimension is dimensionless.
        """
        true_shape = np.shape(self.long_signals)
        output = elephant.signal_processing.hilbert(
            self.long_signals, padding='nextpow')
        self.assertEqual(np.shape(output), true_shape)
        self.assertEqual(output.units, pq.dimensionless)
        output = elephant.signal_processing.hilbert(
            self.long_signals, padding=16384)
        self.assertEqual(np.shape(output), true_shape)
        self.assertEqual(output.units, pq.dimensionless)

    def test_hilbert_array_annotations(self):
        output = elephant.signal_processing.hilbert(self.long_signals,
                                                    padding='nextpow')
        # Test if array_annotations are preserved
        self.assertSetEqual(set(output.array_annotations.keys()), {"my_list"})
        assert_array_equal(output.array_annotations['my_list'],
                           self.long_signals.array_annotations['my_list'])

    def test_hilbert_theoretical_long_signals(self):
        """
        Tests the output of the hilbert function with regard to amplitude and
        phase of long test signals
        """
        # Performing test using all pad types
        for padding in ['nextpow', 'none', 16384]:
            h = elephant.signal_processing.hilbert(
                self.long_signals, padding=padding)

            phase = np.angle(h.magnitude)
            amplitude = np.abs(h.magnitude)
            real_value = np.real(h.magnitude)

            # The real part should be equal to the original long_signals
            assert_array_almost_equal(
                real_value,
                self.long_signals.magnitude,
                decimal=14)

            # Test only in the middle half of the array (border effects)
            ind1 = int(len(h.times) / 4)
            ind2 = int(3 * len(h.times) / 4)

            # Calculate difference in phase between signal and original phase
            # and use smaller of any two phase differences
            phasediff = np.abs(phase[ind1:ind2, :] - self.phase[ind1:ind2, :])
            phasediff[phasediff >= np.pi] = \
                2 * np.pi - phasediff[phasediff >= np.pi]

            # Calculate difference in amplitude between signal and original
            # amplitude
            amplitudediff = \
                amplitude[ind1:ind2, :] - self.amplitude[ind1:ind2, :]
            #
            assert_allclose(phasediff, 0, atol=0.1)
            assert_allclose(amplitudediff, 0, atol=0.5)

    def test_hilbert_theoretical_one_period(self):
        """
        Tests the output of the hilbert function with regard to amplitude and
        phase of a short signal covering one cycle (more accurate estimate).

        This unit test is adapted from the scipy library of the hilbert()
        function.
        """

        # Precision of testing
        decimal = 14

        # Performing test using both pad types
        for padding in ['nextpow', 'none', 512]:
            h = elephant.signal_processing.hilbert(
                self.one_period, padding=padding)

            amplitude = np.abs(h.magnitude)
            phase = np.angle(h.magnitude)
            real_value = np.real(h.magnitude)

            # The real part should be equal to the original long_signals:
            assert_array_almost_equal(
                real_value,
                self.one_period.magnitude,
                decimal=decimal)

            # The absolute value should be 1 everywhere, for this input:
            assert_array_almost_equal(
                amplitude,
                np.ones(self.one_period.magnitude.shape),
                decimal=decimal)

            # For the 'slow' sine - the phase should go from -pi/2 to pi/2 in
            # the first 256 bins:
            assert_array_almost_equal(
                phase[:256, 0],
                np.arange(-np.pi / 2, np.pi / 2, np.pi / 256),
                decimal=decimal)
            # For the 'slow' cosine - the phase should go from 0 to pi in the
            # same interval:
            assert_array_almost_equal(
                phase[:256, 1],
                np.arange(0, np.pi, np.pi / 256),
                decimal=decimal)
            # The 'fast' sine should make this phase transition in half the
            # time:
            assert_array_almost_equal(
                phase[:128, 2],
                np.arange(-np.pi / 2, np.pi / 2, np.pi / 128),
                decimal=decimal)
            # The 'fast' cosine should make this phase transition in half the
            # time:
            assert_array_almost_equal(
                phase[:128, 3],
                np.arange(0, np.pi, np.pi / 128),
                decimal=decimal)


class WaveletTestCase(unittest.TestCase):
    def setUp(self):
        # generate a 10-sec test data of pure 50 Hz cosine wave
        self.fs = 1000.0
        self.times = np.arange(0, 10.0, 1 / self.fs)
        self.test_freq1 = 50.0
        self.test_freq2 = 60.0
        self.test_data1 = np.cos(2 * np.pi * self.test_freq1 * self.times)
        self.test_data2 = np.sin(2 * np.pi * self.test_freq2 * self.times)
        self.test_data_arr = np.vstack([self.test_data1, self.test_data2])
        self.test_data = neo.AnalogSignal(
            self.test_data_arr.T * pq.mV, t_start=self.times[0] * pq.s,
            t_stop=self.times[-1] * pq.s, sampling_period=(1 / self.fs) * pq.s)
        self.true_phase1 = np.angle(
            self.test_data1 +
            1j *
            np.sin(
                2 *
                np.pi *
                self.test_freq1 *
                self.times))
        self.true_phase2 = np.angle(
            self.test_data2 -
            1j *
            np.cos(
                2 *
                np.pi *
                self.test_freq2 *
                self.times))
        self.wt_freqs = [10, 20, 30]

    def test_wavelet_errors(self):
        """
        Tests if errors are raised as expected.
        """
        # too high center frequency
        kwds = {'signal': self.test_data, 'frequency': self.fs / 2}
        self.assertRaises(
            ValueError, elephant.signal_processing.wavelet_transform, **kwds)
        kwds = {
            'signal': self.test_data_arr,
            'frequency': self.fs / 2,
            'sampling_frequency': self.fs}
        self.assertRaises(
            ValueError, elephant.signal_processing.wavelet_transform, **kwds)

        # too high center frequency in a list
        kwds = {'signal': self.test_data,
                'frequency': [self.fs / 10, self.fs / 2]}
        self.assertRaises(
            ValueError, elephant.signal_processing.wavelet_transform, **kwds)
        kwds = {'signal': self.test_data_arr,
                'frequency': [self.fs / 10, self.fs / 2],
                'sampling_frequency': self.fs}
        self.assertRaises(
            ValueError, elephant.signal_processing.wavelet_transform, **kwds)

        # nco is not positive
        kwds = {'signal': self.test_data, 'frequency': self.fs / 10,
                'n_cycles': 0}
        self.assertRaises(
            ValueError, elephant.signal_processing.wavelet_transform, **kwds)

    def test_wavelet_io(self):
        """
        Tests the data type and data shape of the output is consistent with
        that of the input, and also test the consistency between the outputs
        of different types
        """
        # check the shape of the result array
        # --- case of single center frequency
        wt = elephant.signal_processing.wavelet_transform(self.test_data,
                                                          self.fs / 10)
        self.assertTrue(wt.ndim == self.test_data.ndim)
        self.assertTrue(wt.shape[0] == self.test_data.shape[0])  # time axis
        self.assertTrue(wt.shape[1] == self.test_data.shape[1])  # channel axis

        wt_arr = elephant.signal_processing.wavelet_transform(
            self.test_data_arr, self.fs / 10, sampling_frequency=self.fs)
        self.assertTrue(wt_arr.ndim == self.test_data.ndim)
        # channel axis
        self.assertTrue(wt_arr.shape[0] == self.test_data_arr.shape[0])
        # time axis
        self.assertTrue(wt_arr.shape[1] == self.test_data_arr.shape[1])

        wt_arr1d = elephant.signal_processing.wavelet_transform(
            self.test_data1, self.fs / 10, sampling_frequency=self.fs)
        self.assertTrue(wt_arr1d.ndim == self.test_data1.ndim)
        # time axis
        self.assertTrue(wt_arr1d.shape[0] == self.test_data1.shape[0])

        # --- case of multiple center frequencies
        wt = elephant.signal_processing.wavelet_transform(
            self.test_data, self.wt_freqs)
        self.assertTrue(wt.ndim == self.test_data.ndim + 1)
        self.assertTrue(wt.shape[0] == self.test_data.shape[0])  # time axis
        self.assertTrue(wt.shape[1] == self.test_data.shape[1])  # channel axis
        self.assertTrue(wt.shape[2] == len(self.wt_freqs))  # frequency axis

        wt_arr = elephant.signal_processing.wavelet_transform(
            self.test_data_arr, self.wt_freqs, sampling_frequency=self.fs)
        self.assertTrue(wt_arr.ndim == self.test_data_arr.ndim + 1)
        # channel axis
        self.assertTrue(wt_arr.shape[0] == self.test_data_arr.shape[0])
        # frequency axis
        self.assertTrue(wt_arr.shape[1] == len(self.wt_freqs))
        # time axis
        self.assertTrue(wt_arr.shape[2] == self.test_data_arr.shape[1])

        wt_arr1d = elephant.signal_processing.wavelet_transform(
            self.test_data1, self.wt_freqs, sampling_frequency=self.fs)
        self.assertTrue(wt_arr1d.ndim == self.test_data1.ndim + 1)
        # frequency axis
        self.assertTrue(wt_arr1d.shape[0] == len(self.wt_freqs))
        # time axis
        self.assertTrue(wt_arr1d.shape[1] == self.test_data1.shape[0])

        # check that the result does not depend on data type
        self.assertTrue(np.all(wt[:, 0, :] == wt_arr[0, :, :].T))  # channel 0
        self.assertTrue(np.all(wt[:, 1, :] == wt_arr[1, :, :].T))  # channel 1

        # check the data contents in the case where freq is given as a list
        # Note: there seems to be a bug in np.fft since NumPy 1.14.1, which
        # causes that the values of wt_1freq[:, 0] and wt_3freqs[:, 0, 0] are
        # not exactly equal, even though they use the same center frequency for
        # wavelet transform (in NumPy 1.13.1, they become identical). Here we
        # only check that they are almost equal.
        wt_1freq = elephant.signal_processing.wavelet_transform(
            self.test_data, self.wt_freqs[0])
        wt_3freqs = elephant.signal_processing.wavelet_transform(
            self.test_data, self.wt_freqs)
        assert_array_almost_equal(wt_1freq[:, 0], wt_3freqs[:, 0, 0],
                                  decimal=12)

    def test_wavelet_amplitude(self):
        """
        Tests amplitude properties of the obtained wavelet transform
        """
        # check that the amplitude of WT of a sinusoid is (almost) constant
        wt = elephant.signal_processing.wavelet_transform(self.test_data,
                                                          self.test_freq1)
        # take a middle segment in order to avoid edge effects
        amp = np.abs(wt[int(len(wt) / 3):int(len(wt) // 3 * 2), 0])
        mean_amp = amp.mean()
        assert_array_almost_equal((amp - mean_amp) / mean_amp,
                                  np.zeros_like(amp), decimal=6)

        # check that the amplitude of WT is (almost) zero when center frequency
        # is considerably different from signal frequency
        wt_low = elephant.signal_processing.wavelet_transform(
            self.test_data, self.test_freq1 / 10)
        amp_low = np.abs(wt_low[int(len(wt) / 3):int(len(wt) // 3 * 2), 0])
        assert_array_almost_equal(amp_low, np.zeros_like(amp), decimal=6)

        # check that zero padding hardly affect the result
        wt_padded = elephant.signal_processing.wavelet_transform(
            self.test_data, self.test_freq1, zero_padding=False)
        amp_padded = np.abs(
            wt_padded[int(len(wt) / 3):int(len(wt) // 3 * 2), 0])
        assert_array_almost_equal(amp_padded, amp, decimal=9)

    def test_wavelet_phase(self):
        """
        Tests phase properties of the obtained wavelet transform
        """
        # check that the phase of WT is (almost) same as that of the original
        # sinusoid
        wt = elephant.signal_processing.wavelet_transform(self.test_data,
                                                          self.test_freq1)
        phase = np.angle(wt[int(len(wt) / 3):int(len(wt) // 3 * 2), 0])
        true_phase = self.true_phase1[int(len(wt) / 3):int(len(wt) // 3 * 2)]
        assert_array_almost_equal(np.exp(1j * phase), np.exp(1j * true_phase),
                                  decimal=6)

        # check that zero padding hardly affect the result
        wt_padded = elephant.signal_processing.wavelet_transform(
            self.test_data, self.test_freq1, zero_padding=False)
        phase_padded = np.angle(
            wt_padded[int(len(wt) / 3):int(len(wt) // 3 * 2), 0])
        assert_array_almost_equal(
            np.exp(
                1j * phase_padded),
            np.exp(
                1j * phase),
            decimal=9)


class DerivativeTestCase(unittest.TestCase):

    def setUp(self):
        self.fs = 1000.0
        self.tmin = 0.0
        self.tmax = 10.0
        self.times = np.arange(self.tmin, self.tmax, 1 / self.fs)
        self.test_data1 = np.cos(2 * np.pi * self.times)
        self.test_data2 = np.vstack(
            [np.cos(2 * np.pi * self.times), np.sin(2 * np.pi * self.times)]).T
        self.test_signal1 = neo.AnalogSignal(
            self.test_data1 * pq.mV, t_start=self.times[0] * pq.s,
            t_stop=self.times[-1] * pq.s, sampling_period=(1 / self.fs) * pq.s)
        self.test_signal2 = neo.AnalogSignal(
            self.test_data2 * pq.mV, t_start=self.times[0] * pq.s,
            t_stop=self.times[-1] * pq.s, sampling_period=(1 / self.fs) * pq.s)

    def test_derivative_invalid_signal(self):
        """Test derivative on non-AnalogSignal"""
        kwds = {'signal': np.arange(5)}
        self.assertRaises(
            TypeError, elephant.signal_processing.derivative, **kwds)

    def test_derivative_units(self):
        """Test derivative returns AnalogSignal with correct units"""
        derivative = elephant.signal_processing.derivative(
            self.test_signal1)
        self.assertTrue(isinstance(derivative, neo.AnalogSignal))
        self.assertEqual(
            derivative.units,
            self.test_signal1.units / self.test_signal1.times.units)

    def test_derivative_times(self):
        """Test derivative returns AnalogSignal with correct times"""
        derivative = elephant.signal_processing.derivative(
            self.test_signal1)
        self.assertTrue(isinstance(derivative, neo.AnalogSignal))

        # test that sampling period is correct
        self.assertEqual(
            derivative.sampling_period,
            1 / self.fs * self.test_signal1.times.units)

        # test that all times are correct
        target_times = self.times[:-1] * self.test_signal1.times.units \
            + derivative.sampling_period / 2
        assert_array_almost_equal(derivative.times, target_times)

        # test that t_start and t_stop are correct
        self.assertEqual(derivative.t_start, target_times[0])
        assert_array_almost_equal(
            derivative.t_stop,
            target_times[-1] + derivative.sampling_period)

    def test_derivative_values(self):
        """Test derivative returns AnalogSignal with correct values"""
        derivative1 = elephant.signal_processing.derivative(
            self.test_signal1)
        derivative2 = elephant.signal_processing.derivative(
            self.test_signal2)
        self.assertTrue(isinstance(derivative1, neo.AnalogSignal))
        self.assertTrue(isinstance(derivative2, neo.AnalogSignal))

        # single channel
        assert_array_almost_equal(
            derivative1.magnitude,
            np.vstack([np.diff(self.test_data1)]).T / (1 / self.fs))

        # multi channel
        assert_array_almost_equal(derivative2.magnitude, np.vstack([
            np.diff(self.test_data2[:, 0]),
            np.diff(self.test_data2[:, 1])]).T / (1 / self.fs))


class RAUCTestCase(unittest.TestCase):

    def setUp(self):
        self.fs = 1000.0
        self.tmin = 0.0
        self.tmax = 10.0
        self.times = np.arange(self.tmin, self.tmax, 1 / self.fs)
        self.test_data1 = np.cos(2 * np.pi * self.times)
        self.test_data2 = np.vstack(
            [np.cos(2 * np.pi * self.times), np.sin(2 * np.pi * self.times)]).T
        self.test_signal1 = neo.AnalogSignal(
            self.test_data1 * pq.mV, t_start=self.times[0] * pq.s,
            t_stop=self.times[-1] * pq.s, sampling_period=(1 / self.fs) * pq.s)
        self.test_signal2 = neo.AnalogSignal(
            self.test_data2 * pq.mV, t_start=self.times[0] * pq.s,
            t_stop=self.times[-1] * pq.s, sampling_period=(1 / self.fs) * pq.s)

    def test_rauc_invalid_signal(self):
        """Test rauc on non-AnalogSignal"""
        kwds = {'signal': np.arange(5)}
        self.assertRaises(
            ValueError, elephant.signal_processing.rauc, **kwds)

    def test_rauc_invalid_bin_duration(self):
        """Test rauc on bad bin duration"""
        kwds = {'signal': self.test_signal1, 'bin_duration': 'bad'}
        self.assertRaises(
            ValueError, elephant.signal_processing.rauc, **kwds)

    def test_rauc_invalid_baseline(self):
        """Test rauc on bad baseline"""
        kwds = {'signal': self.test_signal1, 'baseline': 'bad'}
        self.assertRaises(
            ValueError, elephant.signal_processing.rauc, **kwds)

    def test_rauc_units(self):
        """Test rauc returns Quantity or AnalogSignal with correct units"""

        # test that single-bin result is Quantity with correct units
        rauc = elephant.signal_processing.rauc(
            self.test_signal1)
        self.assertTrue(isinstance(rauc, pq.Quantity))
        self.assertEqual(
            rauc.units,
            self.test_signal1.units * self.test_signal1.times.units)

        # test that multi-bin result is AnalogSignal with correct units
        rauc_arr = elephant.signal_processing.rauc(
            self.test_signal1, bin_duration=1 * pq.s)
        self.assertTrue(isinstance(rauc_arr, neo.AnalogSignal))
        self.assertEqual(
            rauc_arr.units,
            self.test_signal1.units * self.test_signal1.times.units)

    def test_rauc_times_without_overextending_bin(self):
        """Test rauc returns correct times when signal is binned evenly"""

        bin_duration = 1 * pq.s  # results in all bin centers < original t_stop
        rauc_arr = elephant.signal_processing.rauc(
            self.test_signal1, bin_duration=bin_duration)
        self.assertTrue(isinstance(rauc_arr, neo.AnalogSignal))

        # test that sampling period is correct
        self.assertEqual(rauc_arr.sampling_period, bin_duration)

        # test that all times are correct
        target_times = np.arange(self.tmin,
                                 self.tmax,
                                 bin_duration.magnitude) \
            * bin_duration.units + bin_duration / 2
        assert_array_almost_equal(rauc_arr.times, target_times)

        # test that t_start and t_stop are correct
        self.assertEqual(rauc_arr.t_start, target_times[0])
        assert_array_almost_equal(
            rauc_arr.t_stop,
            target_times[-1] + bin_duration)

    def test_rauc_times_with_overextending_bin(self):
        """Test rauc returns correct times when signal is NOT binned evenly"""
        # results in one bin center > original t_stop
        bin_duration = 0.99 * pq.s
        rauc_arr = elephant.signal_processing.rauc(
            self.test_signal1, bin_duration=bin_duration)
        self.assertTrue(isinstance(rauc_arr, neo.AnalogSignal))

        # test that sampling period is correct
        self.assertEqual(rauc_arr.sampling_period, bin_duration)

        # test that all times are correct
        target_times = np.arange(self.tmin,
                                 self.tmax,
                                 bin_duration.magnitude) \
            * bin_duration.units + bin_duration / 2
        assert_array_almost_equal(rauc_arr.times, target_times)

        # test that t_start and t_stop are correct
        self.assertEqual(rauc_arr.t_start, target_times[0])
        assert_array_almost_equal(
            rauc_arr.t_stop,
            target_times[-1] + bin_duration)

    def test_rauc_values_one_bin(self):
        """Test rauc returns correct values when there is just one bin"""
        rauc1 = elephant.signal_processing.rauc(
            self.test_signal1)
        rauc2 = elephant.signal_processing.rauc(
            self.test_signal2)
        self.assertTrue(isinstance(rauc1, pq.Quantity))
        self.assertTrue(isinstance(rauc2, pq.Quantity))

        # single channel
        assert_array_almost_equal(
            rauc1.magnitude,
            np.array([6.36517679]))

        # multi channel
        assert_array_almost_equal(
            rauc2.magnitude,
            np.array([6.36517679, 6.36617364]))

    def test_rauc_values_multi_bin(self):
        """Test rauc returns correct values when there are multiple bins"""
        rauc_arr1 = elephant.signal_processing.rauc(
            self.test_signal1, bin_duration=0.99 * pq.s)
        rauc_arr2 = elephant.signal_processing.rauc(
            self.test_signal2, bin_duration=0.99 * pq.s)
        self.assertTrue(isinstance(rauc_arr1, neo.AnalogSignal))
        self.assertTrue(isinstance(rauc_arr2, neo.AnalogSignal))

        # single channel
        assert_array_almost_equal(rauc_arr1.magnitude, np.array([
            [0.62562647],
            [0.62567202],
            [0.62576076],
            [0.62589236],
            [0.62606628],
            [0.62628184],
            [0.62653819],
            [0.62683432],
            [0.62716907],
            [0.62754110],
            [0.09304862]]))

        # multi channel
        assert_array_almost_equal(rauc_arr2.magnitude, np.array([
            [0.62562647, 0.63623770],
            [0.62567202, 0.63554830],
            [0.62576076, 0.63486313],
            [0.62589236, 0.63418488],
            [0.62606628, 0.63351623],
            [0.62628184, 0.63285983],
            [0.62653819, 0.63221825],
            [0.62683432, 0.63159403],
            [0.62716907, 0.63098964],
            [0.62754110, 0.63040747],
            [0.09304862, 0.03039579]]))

    def test_rauc_mean_baseline(self):
        """Test rauc returns correct values when baseline='mean' is given"""
        rauc1 = elephant.signal_processing.rauc(
            self.test_signal1, baseline='mean')
        rauc2 = elephant.signal_processing.rauc(
            self.test_signal2, baseline='mean')
        self.assertTrue(isinstance(rauc1, pq.Quantity))
        self.assertTrue(isinstance(rauc2, pq.Quantity))

        # single channel
        assert_array_almost_equal(
            rauc1.magnitude,
            np.array([6.36517679]))

        # multi channel
        assert_array_almost_equal(
            rauc2.magnitude,
            np.array([6.36517679, 6.36617364]))

    def test_rauc_median_baseline(self):
        """Test rauc returns correct values when baseline='median' is given"""
        rauc1 = elephant.signal_processing.rauc(
            self.test_signal1, baseline='median')
        rauc2 = elephant.signal_processing.rauc(
            self.test_signal2, baseline='median')
        self.assertTrue(isinstance(rauc1, pq.Quantity))
        self.assertTrue(isinstance(rauc2, pq.Quantity))

        # single channel
        assert_array_almost_equal(
            rauc1.magnitude,
            np.array([6.36517679]))

        # multi channel
        assert_array_almost_equal(
            rauc2.magnitude,
            np.array([6.36517679, 6.36617364]))

    def test_rauc_arbitrary_baseline(self):
        """Test rauc returns correct values when arbitrary baseline is given"""
        rauc1 = elephant.signal_processing.rauc(
            self.test_signal1, baseline=0.123 * pq.mV)
        rauc2 = elephant.signal_processing.rauc(
            self.test_signal2, baseline=0.123 * pq.mV)
        self.assertTrue(isinstance(rauc1, pq.Quantity))
        self.assertTrue(isinstance(rauc2, pq.Quantity))

        # single channel
        assert_array_almost_equal(
            rauc1.magnitude,
            np.array([6.41354725]))

        # multi channel
        assert_array_almost_equal(
            rauc2.magnitude,
            np.array([6.41354725, 6.41429810]))

    def test_rauc_time_slice(self):
        """Test rauc returns correct values when t_start, t_stop are given"""
        rauc1 = elephant.signal_processing.rauc(
            self.test_signal1, t_start=0.123 * pq.s, t_stop=0.456 * pq.s)
        rauc2 = elephant.signal_processing.rauc(
            self.test_signal2, t_start=0.123 * pq.s, t_stop=0.456 * pq.s)
        self.assertTrue(isinstance(rauc1, pq.Quantity))
        self.assertTrue(isinstance(rauc2, pq.Quantity))

        # single channel
        assert_array_almost_equal(
            rauc1.magnitude,
            np.array([0.16279006]))

        # multi channel
        assert_array_almost_equal(
            rauc2.magnitude,
            np.array([0.16279006, 0.26677944]))


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from unittest.mock import patch
from pathlib import Path
import urllib

from elephant.datasets import download_datasets


class TestDownloadDatasets(unittest.TestCase):
    @patch.dict(os.environ, {'ELEPHANT_DATA_LOCATION': '/valid/path'}, clear=True)
    @patch('os.path.exists', return_value=True)
    def test_valid_path(self, mock_exists):
        repo_path = 'some/repo/path'
        expected = Path('/valid/path/some/repo/path')
        result = download_datasets(repo_path)
        self.assertEqual(result, expected)

    @patch.dict(os.environ, {'ELEPHANT_DATA_LOCATION': 'http://valid.url'}, clear=True)
    @patch('os.path.exists', return_value=False)
    def test_valid_url(self, mock_exists):
        repo_path = 'some/repo/path'
        self.assertRaises(urllib.error.URLError, download_datasets, repo_path)

    @patch.dict(os.environ, {'ELEPHANT_DATA_LOCATION': 'invalid_path_or_url'}, clear=True)
    @patch('os.path.exists', return_value=False)
    def test_invalid_value(self, mock_exists):
        repo_path = 'some/repo/path'
        with self.assertRaises(ValueError) as cm:
            download_datasets(repo_path)
        self.assertIn("invalid_path_or_url", str(cm.exception))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Unit tests for the kCSD methods

This was written by :
Chaitanya Chintaluri,
Laboratory of Neuroinformatics,
Nencki Institute of Exprimental Biology, Warsaw.

:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest
import numpy as np
import quantities as pq
from elephant import current_source_density as csd
import elephant.current_source_density_src.utility_functions as utils
from elephant.current_source_density import generate_lfp

available_1d = ['StandardCSD', 'DeltaiCSD', 'StepiCSD', 'SplineiCSD', 'KCSD1D']
available_2d = ['KCSD2D', 'MoIKCSD']
available_3d = ['KCSD3D']
kernel_methods = ['KCSD1D', 'KCSD2D', 'KCSD3D', 'MoIKCSD']
icsd_methods = ['DeltaiCSD', 'StepiCSD', 'SplineiCSD']
py_iCSD_toolbox = ['StandardCSD', 'DeltaiCSD', 'StepiCSD', 'SplineiCSD']


class LFP_TestCase(unittest.TestCase):
    def test_lfp1d_electrodes(self):
        ele_pos = utils.generate_electrodes(dim=1).reshape(5, 1)
        lfp = csd.generate_lfp(utils.gauss_1d_dipole, ele_pos)
        self.assertEqual(ele_pos.shape[1], 1)
        self.assertEqual(ele_pos.shape[0], lfp.shape[1])

    def test_lfp2d_electrodes(self):
        ele_pos = utils.generate_electrodes(dim=2)
        xx_ele, yy_ele = ele_pos
        lfp = csd.generate_lfp(utils.large_source_2D, xx_ele, yy_ele)
        self.assertEqual(len(ele_pos), 2)
        self.assertEqual(xx_ele.shape[0], lfp.shape[1])

    def test_lfp3d_electrodes(self):
        ele_pos = utils.generate_electrodes(dim=3, res=3)
        xx_ele, yy_ele, zz_ele = ele_pos
        lfp = csd.generate_lfp(utils.gauss_3d_dipole, xx_ele, yy_ele, zz_ele)
        self.assertEqual(len(ele_pos), 3)
        self.assertEqual(xx_ele.shape[0], lfp.shape[1])


class CSD1D_TestCase(unittest.TestCase):
    def setUp(self):
        self.ele_pos = utils.generate_electrodes(dim=1).reshape(5, 1)
        self.lfp = csd.generate_lfp(utils.gauss_1d_dipole, self.ele_pos)
        self.csd_method = csd.estimate_csd

        # Input dictionaries for each method
        self.params = {'DeltaiCSD': {'sigma_top': 0. * pq.S / pq.m,
                                     'diam': 500E-6 * pq.m},
                       'StepiCSD': {'sigma_top': 0. * pq.S / pq.m,
                                    'tol': 1E-12,
                                    'diam': 500E-6 * pq.m},
                       'SplineiCSD': {'sigma_top': 0. * pq.S / pq.m,
                                      'num_steps': 201, 'tol': 1E-12,
                                      'diam': 500E-6 * pq.m},
                       'StandardCSD': {}, 'KCSD1D': {'h': 50.,
                                                     'Rs': np.array(
                                                        (0.1, 0.25, 0.5))}}

    def test_validate_inputs(self):
        self.assertRaises(TypeError, self.csd_method, lfp=[[1], [2], [3]])
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          coordinates=self.ele_pos * pq.mm)
        # inconsistent number of electrodes
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          coordinates=[1, 2, 3, 4] * pq.mm,
                          method='StandardCSD')
        # bad method name
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          method='InvalidMethodName')
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          method='KCSD2D')
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          method='KCSD3D')

    def test_inputs_standardcsd(self):
        method = 'StandardCSD'
        result = self.csd_method(self.lfp, method=method)
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.shape[0], 1)

    def test_inputs_deltasplineicsd(self):
        methods = ['DeltaiCSD', 'SplineiCSD']
        for method in methods:
            self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                              method=method)
            result = self.csd_method(self.lfp, method=method,
                                     **self.params[method])
            self.assertEqual(result.t_start, 0.0 * pq.s)
            self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
            self.assertEqual(result.times.shape[0], 1)

    def test_inputs_stepicsd(self):
        method = 'StepiCSD'
        self.assertRaises(ValueError, self.csd_method, lfp=self.lfp,
                          method=method)
        self.assertRaises(AssertionError, self.csd_method, lfp=self.lfp,
                          method=method, **self.params[method])
        self.params['StepiCSD'].update({'h': np.ones(5) * 100E-6 * pq.m})
        result = self.csd_method(self.lfp, method=method,
                                 **self.params[method])
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(result.times.shape[0], 1)

    def test_inuts_kcsd(self):
        method = 'KCSD1D'
        result = self.csd_method(self.lfp, method=method,
                                 **self.params[method])
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(len(result.times), 1)


class CSD2D_TestCase(unittest.TestCase):
    def setUp(self):
        xx_ele, yy_ele = utils.generate_electrodes(dim=2)
        self.lfp = csd.generate_lfp(utils.large_source_2D, xx_ele, yy_ele)
        self.params = {'KCSD2D': {'sigma': 1., 'Rs': np.array(
            (0.1, 0.25, 0.5))}}  # Input dictionaries for each method

    def test_kcsd2d_init(self):
        method = 'KCSD2D'
        result = csd.estimate_csd(lfp=self.lfp, method=method,
                                  **self.params[method])
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(len(result.times), 1)


class CSD3D_TestCase(unittest.TestCase):
    def setUp(self):
        xx_ele, yy_ele, zz_ele = utils.generate_electrodes(dim=3)
        self.lfp = csd.generate_lfp(utils.gauss_3d_dipole,
                                    xx_ele, yy_ele, zz_ele)
        self.params = {'KCSD3D': {'gdx': 0.1, 'gdy': 0.1, 'gdz': 0.1,
                                  'src_type': 'step',
                                  'Rs': np.array((0.1, 0.25, 0.5))}}

    def test_kcsd2d_init(self):
        method = 'KCSD3D'
        result = csd.estimate_csd(lfp=self.lfp, method=method,
                                  **self.params[method])
        self.assertEqual(result.t_start, 0.0 * pq.s)
        self.assertEqual(result.sampling_rate, 1000 * pq.Hz)
        self.assertEqual(len(result.times), 1)


class GenerateLfpTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.one_dimensional = np.linspace(0, 10, 2304)
        cls.two_dimensional = np.linspace(0, 10, 2304
                                          ).reshape(2304, 1)

    def test_generate_lfp_one_dimensional_array(self):
        """
        Regression test for Issue #546,
        see: https://github.com/NeuralEnsemble/elephant/issues/546
        """
        # this should raise NOT an error
        generate_lfp(utils.gauss_1d_dipole, self.one_dimensional)

    def test_generate_lfp_two_dimensional_array(self):
        """
        Regression test for Issue #546,
        see: https://github.com/NeuralEnsemble/elephant/issues/546
        """
        # this should NOT raise an error
        generate_lfp(utils.gauss_1d_dipole, self.two_dimensional)


if __name__ == '__main__':
    unittest.main()



import mpi4py.futures
from mpi4py import MPI

from .parallel import ProcessPoolExecutor


class MPIPoolExecutor(ProcessPoolExecutor):
    """
    The `MPIPoolExecutor` class uses a pool of MPI processes to execute calls
    asynchronously.

    `MPIPoolExecutor` is recommended to use on cluster nodes which support
    MPI-2 protocol.

    Notes
    -----
    `-m mpi4py.futures` command line option is needed to execute Python scripts
    with MPI:

    .. code-block:: sh

       mpiexec -n numprocs python -m mpi4py.futures pyfile [arg] ...

    For more information of how to launch MPI processes in Python refer to
    https://mpi4py.readthedocs.io/en/stable/mpi4py.futures.html#command-line
    """

    def _create_executor(self):
        return mpi4py.futures.MPIPoolExecutor(self.max_workers)

    def _extra_repr(self):
        extra_old = super(MPIPoolExecutor, self)._extra_repr()
        info = dict(MPI.INFO_ENV)
        return "{old}, {mpi}".format(old=extra_old, mpi=info)


class MPICommExecutor(MPIPoolExecutor):
    """
    Legacy MPI-1 implementation for cluster nodes which do not support MPI-2
    protocol.

    Parameters
    ----------
    comm : MPI.Intracomm or None
        MPI (intra)communicator. If None, set to `MPI.COMM_WORLD`.
        Default: None
    root : int
        Designated master process.
        Default: 0

    Notes
    -----
    `-m mpi4py.futures` command line option is needed to execute Python scripts
    with MPI:

    .. code-block:: sh

       mpiexec -n numprocs python -m mpi4py.futures pyfile [arg] ...

    For more information of how to launch MPI processes in Python refer to
    https://mpi4py.readthedocs.io/en/stable/mpi4py.futures.html#command-line
    """
    def __init__(self, comm=None, root=0):
        super(MPICommExecutor, self).__init__(max_workers=None)
        if comm is None:
            comm = MPI.COMM_WORLD
        self.comm = comm
        self.root = root

    def _create_executor(self):
        return mpi4py.futures.MPICommExecutor(comm=self.comm, root=self.root)



# -*- coding: utf-8 -*-
"""
Functions to calculate spike-triggered average and spike-field coherence of
analog signals.

.. autosummary::
    :toctree: _toctree/sta

    spike_triggered_average
    spike_field_coherence

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import warnings

import numpy as np
import quantities as pq
import scipy.signal
from neo.core import AnalogSignal, SpikeTrain

from .conversion import BinnedSpikeTrain

__all__ = [
    "spike_triggered_average",
    "spike_field_coherence"
]


def spike_triggered_average(signal, spiketrains, window):
    """
    Calculates the spike-triggered averages of analog signals in a time window
    relative to the spike times of a corresponding spiketrain for multiple
    signals each. The function receives n analog signals and either one or
    n spiketrains. In case it is one spiketrain this one is muliplied n-fold
    and used for each of the n analog signals.

    Parameters
    ----------
    signal : neo AnalogSignal object
        'signal' contains n analog signals.
    spiketrains : one SpikeTrain or one numpy ndarray or a list of n of either of these.
        'spiketrains' contains the times of the spikes in the spiketrains.
    window : tuple of 2 Quantity objects with dimensions of time.
        'window' is the start time and the stop time, relative to a spike, of
        the time interval for signal averaging.
        If the window size is not a multiple of the sampling interval of the
        signal the window will be extended to the next multiple.

    Returns
    -------
    result_sta : neo AnalogSignal object
        'result_sta' contains the spike-triggered averages of each of the
        analog signals with respect to the spikes in the corresponding
        spiketrains. The length of 'result_sta' is calculated as the number
        of bins from the given start and stop time of the averaging interval
        and the sampling rate of the analog signal. If for an analog signal
        no spike was either given or all given spikes had to be ignored
        because of a too large averaging interval, the corresponding returned
        analog signal has all entries as nan. The number of used spikes and
        unused spikes for each analog signal are returned as annotations to
        the returned AnalogSignal object.

    Examples
    --------
    >>> import neo

    >>> signal = neo.AnalogSignal(np.array([signal1, signal2]).T, units='mV',
    ...                                sampling_rate=10/ms) # doctest: +SKIP
    >>> stavg = spike_triggered_average(signal, [spiketrain1, spiketrain2],
    ...                                 (-5 * ms, 10 * ms)) # doctest: +SKIP

    """

    # checking compatibility of data and data types
    # window_starttime: time to specify the start time of the averaging
    # interval relative to a spike
    # window_stoptime: time to specify the stop time of the averaging
    # interval relative to a spike
    window_starttime, window_stoptime = window
    if not (isinstance(window_starttime, pq.quantity.Quantity) and
            window_starttime.dimensionality.simplified ==
            pq.Quantity(1, "s").dimensionality):
        raise TypeError("The start time of the window (window[0]) "
                        "must be a time quantity.")
    if not (isinstance(window_stoptime, pq.quantity.Quantity) and
            window_stoptime.dimensionality.simplified ==
            pq.Quantity(1, "s").dimensionality):
        raise TypeError("The stop time of the window (window[1]) "
                        "must be a time quantity.")
    if window_stoptime <= window_starttime:
        raise ValueError("The start time of the window (window[0]) must be "
                         "earlier than the stop time of the window (window[1]).")

    # checks on signal
    if not isinstance(signal, AnalogSignal):
        raise TypeError(
            "Signal must be an AnalogSignal, not %s." % type(signal))
    if len(signal.shape) > 1:
        # num_signals: number of analog signals
        num_signals = signal.shape[1]
    else:
        raise ValueError("Empty analog signal, hence no averaging possible.")
    if window_stoptime - window_starttime > signal.t_stop - signal.t_start:
        raise ValueError("The chosen time window is larger than the "
                         "time duration of the signal.")

    # spiketrains type check
    if isinstance(spiketrains, (np.ndarray, SpikeTrain)):
        spiketrains = [spiketrains]
    elif isinstance(spiketrains, list):
        for st in spiketrains:
            if not isinstance(st, (np.ndarray, SpikeTrain)):
                raise TypeError(
                    "spiketrains must be a SpikeTrain, a numpy ndarray, or a "
                    "list of one of those, not %s." % type(spiketrains))
    else:
        raise TypeError(
            "spiketrains must be a SpikeTrain, a numpy ndarray, or a list of "
            "one of those, not %s." % type(spiketrains))

    # multiplying spiketrain in case only a single spiketrain is given
    if len(spiketrains) == 1 and num_signals != 1:
        template = spiketrains[0]
        spiketrains = []
        for i in range(num_signals):
            spiketrains.append(template)

    # checking for matching numbers of signals and spiketrains
    if num_signals != len(spiketrains):
        raise ValueError(
            "The number of signals and spiketrains has to be the same.")

    # checking the times of signal and spiketrains
    for i in range(num_signals):
        if spiketrains[i].t_start < signal.t_start:
            raise ValueError(
                "The spiketrain indexed by %i starts earlier than "
                "the analog signal." % i)
        if spiketrains[i].t_stop > signal.t_stop:
            raise ValueError(
                "The spiketrain indexed by %i stops later than "
                "the analog signal." % i)

    # *** Main algorithm: ***

    # window_bins: number of bins of the chosen averaging interval
    window_bins = int(np.ceil(((window_stoptime - window_starttime) *
        signal.sampling_rate).simplified))
    # result_sta: array containing finally the spike-triggered averaged signal
    result_sta = AnalogSignal(np.zeros((window_bins, num_signals)),
        sampling_rate=signal.sampling_rate, units=signal.units)
    # setting of correct times of the spike-triggered average
    # relative to the spike
    result_sta.t_start = window_starttime
    used_spikes = np.zeros(num_signals, dtype=int)
    unused_spikes = np.zeros(num_signals, dtype=int)
    total_used_spikes = 0

    for i in range(num_signals):
        # summing over all respective signal intervals around spiketimes
        for spiketime in spiketrains[i]:
            # checks for sufficient signal data around spiketime
            if (spiketime + window_starttime >= signal.t_start and
                    spiketime + window_stoptime <= signal.t_stop):
                # calculating the startbin in the analog signal of the
                # averaging window for spike
                startbin = int(np.floor(((spiketime + window_starttime -
                    signal.t_start) * signal.sampling_rate).simplified))
                # adds the signal in selected interval relative to the spike
                result_sta[:, i] += signal[
                    startbin: startbin + window_bins, i]
                # counting of the used spikes
                used_spikes[i] += 1
            else:
                # counting of the unused spikes
                unused_spikes[i] += 1

        # normalization
        result_sta[:, i] = result_sta[:, i] / used_spikes[i]

        total_used_spikes += used_spikes[i]

    if total_used_spikes == 0:
        warnings.warn(
            "No spike at all was either found or used for averaging")
    result_sta.annotate(used_spikes=used_spikes, unused_spikes=unused_spikes)

    return result_sta


def spike_field_coherence(signal, spiketrain, **kwargs):
    """
    Calculates the spike-field coherence between a analog signal(s) and a
    (binned) spike train.

    The current implementation makes use of scipy.signal.coherence(). Additional
    kwargs will will be directly forwarded to scipy.signal.coherence(),
    except for the axis parameter and the sampling frequency, which will be
    extracted from the input signals.

    The spike_field_coherence function receives an analog signal array and
    either a binned spike train or a spike train containing the original spike
    times. In case of original spike times the spike train is binned according
    to the sampling rate of the analog signal array.

    The AnalogSignal object can contain one or multiple signal traces. In case
    of multiple signal traces, the spike field coherence is calculated
    individually for each signal trace and the spike train.

    Parameters
    ----------
    signal : neo AnalogSignal object
        'signal' contains n analog signals.
    spiketrain : SpikeTrain or BinnedSpikeTrain
        Single spike train to perform the analysis on. The bin_size of the
        binned spike train must match the sampling_rate of signal.
    **kwargs:
        All kwargs are passed to `scipy.signal.coherence()`.

    Returns
    -------
    coherence : complex Quantity array
        contains the coherence values calculated for each analog signal trace
        in combination with the spike train. The first dimension corresponds to
        the frequency, the second to the number of the signal trace.
    frequencies : Quantity array
        contains the frequency values corresponding to the first dimension of
        the 'coherence' array

    Examples
    --------

    Plot the SFC between a regular spike train at 20 Hz, and two sinusoidal
    time series at 20 Hz and 23 Hz, respectively.


    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import quantities as pq
        from quantities import s, ms, mV, Hz, kHz
        import neo, elephant

        t = pq.Quantity(range(10000),units='ms')
        f1, f2 = 20. * Hz, 23. * Hz
        signal = neo.AnalogSignal(np.array([
                                  np.sin(f1 * 2. * np.pi * t.rescale(s)),
                                  np.sin(f2 * 2. * np.pi * t.rescale(s))]).T,
                                  units=pq.mV, sampling_rate=1. * kHz)
        spiketrain = neo.SpikeTrain(
           range(t[0], t[-1], 50), units='ms',
           t_start=t[0], t_stop=t[-1])
        sfc, freqs = elephant.sta.spike_field_coherence(
           signal, spiketrain, window='boxcar')

        plt.plot(freqs, sfc[:,0])
        plt.plot(freqs, sfc[:,1])
        plt.xlabel('Frequency [Hz]')
        plt.ylabel('SFC')
        plt.xlim((0, 60))
        plt.show()

    """

    if not hasattr(scipy.signal, 'coherence'):
        raise AttributeError('scipy.signal.coherence is not available. The sfc '
                             'function uses scipy.signal.coherence for '
                             'the coherence calculation. This function is '
                             'available for scipy version 0.16 or newer. '
                             'Please update you scipy version.')

    # spiketrains type check
    if not isinstance(spiketrain, (SpikeTrain, BinnedSpikeTrain)):
        raise TypeError(
            "spiketrain must be of type SpikeTrain or BinnedSpikeTrain, "
            "not %s." % type(spiketrain))

    # checks on analogsignal
    if not isinstance(signal, AnalogSignal):
        raise TypeError(
            "Signal must be an AnalogSignal, not %s." % type(signal))
    if len(signal.shape) > 1:
        # num_signals: number of individual traces in the analog signal
        num_signals = signal.shape[1]
    elif len(signal.shape) == 1:
        num_signals = 1
    else:
        raise ValueError("Empty analog signal.")
    len_signals = signal.shape[0]

    # bin spiketrain if necessary
    if isinstance(spiketrain, SpikeTrain):
        spiketrain = BinnedSpikeTrain(
            spiketrain, bin_size=signal.sampling_period)

    # check the start and stop times of signal and spike trains
    if spiketrain.t_start < signal.t_start:
        raise ValueError(
            "The spiketrain starts earlier than the analog signal.")
    if spiketrain.t_stop > signal.t_stop:
        raise ValueError(
            "The spiketrain stops later than the analog signal.")

    # check equal time resolution for both signals
    if spiketrain.bin_size != signal.sampling_period:
        raise ValueError(
            "The spiketrain and signal must have a "
            "common sampling frequency / bin_size")

    # calculate how many bins to add on the left of the binned spike train
    delta_t = spiketrain.t_start - signal.t_start
    if delta_t % spiketrain.bin_size == 0:
        left_edge = int((delta_t / spiketrain.bin_size).magnitude)
    else:
        raise ValueError("Incompatible binning of spike train and LFP")
    right_edge = int(left_edge + spiketrain.n_bins)

    # duplicate spike trains
    spiketrain_array = np.zeros((1, len_signals))
    spiketrain_array[0, left_edge:right_edge] = spiketrain.to_array()
    spiketrains_array = np.repeat(spiketrain_array, repeats=num_signals, axis=0).transpose()

    # calculate coherence
    frequencies, sfc = scipy.signal.coherence(
        spiketrains_array, signal.magnitude,
        fs=signal.sampling_rate.rescale('Hz').magnitude,
        axis=0, **kwargs)

    return (pq.Quantity(sfc, units=pq.dimensionless),
            pq.Quantity(frequencies, units=pq.Hz))



# -*- coding: utf-8 -*-
"""
Unit tests for the spike_train_generation module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division

import os
import unittest
import warnings

import neo
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_allclose
import quantities as pq
from scipy.stats import expon, kstest, poisson, variation

from elephant.spike_train_generation import StationaryPoissonProcess, \
    threshold_detection, peak_detection, spike_extraction, \
    AbstractPointProcess, StationaryGammaProcess, StationaryLogNormalProcess, \
    NonStationaryPoissonProcess, NonStationaryGammaProcess, \
    StationaryInverseGaussianProcess, _n_poisson, single_interaction_process, \
    cpp, homogeneous_gamma_process, homogeneous_poisson_process, \
    inhomogeneous_poisson_process, inhomogeneous_gamma_process
from elephant.statistics import isi, instantaneous_rate
from elephant import kernels


def pdiff(a, b):
    """Difference between a and b as a fraction of a

    i.e. abs((a - b)/a)
    """
    return abs((a - b) / a)


class AnalogSignalThresholdDetectionTestCase(unittest.TestCase):

    def setUp(self):
        # Load membrane potential simulated using Brian2
        # according to make_spike_extraction_test_data.py.
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        raw_data_file_loc = os.path.join(
            curr_dir, 'spike_extraction_test_data.txt')
        raw_data = []
        with open(raw_data_file_loc, 'r') as f:
            for x in f.readlines():
                raw_data.append(float(x))
        self.vm = neo.AnalogSignal(
            raw_data, units=pq.V, sampling_period=0.1 * pq.ms)
        self.true_time_stamps = [0.0123, 0.0354, 0.0712, 0.1191, 0.1694,
                                 0.2200, 0.2711] * pq.s

    def test_threshold_detection(self):
        # Test whether spikes are extracted at the correct times from
        # an analog signal.

        spike_train = threshold_detection(self.vm)
        try:
            len(spike_train)
        # Handles an error in Neo related to some zero length
        # spike trains being treated as unsized objects.
        except TypeError:
            warnings.warn(
                ("The spike train may be an unsized object. This may be"
                 " related to an issue in Neo with some zero-length SpikeTrain"
                 " objects. Bypassing this by creating an empty SpikeTrain"
                 " object."))
            spike_train = neo.SpikeTrain([], t_start=spike_train.t_start,
                                         t_stop=spike_train.t_stop,
                                         units=spike_train.units)

        # Does threshold_detection gives the correct number of spikes?
        self.assertEqual(len(spike_train), len(self.true_time_stamps))
        # Does threshold_detection gives the correct times for the spikes?
        try:
            assert_array_almost_equal(spike_train, self.true_time_stamps)
        except AttributeError:  # If numpy version too old to have allclose
            self.assertTrue(np.array_equal(spike_train, self.true_time_stamps))

    def test_peak_detection_threshold(self):
        # Test for empty SpikeTrain when threshold is too high
        result = threshold_detection(self.vm, threshold=30 * pq.mV)
        self.assertEqual(len(result), 0)


class AnalogSignalPeakDetectionTestCase(unittest.TestCase):

    def setUp(self):
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        raw_data_file_loc = os.path.join(
            curr_dir, 'spike_extraction_test_data.txt')
        raw_data = []
        with open(raw_data_file_loc, 'r') as f:
            for x in f.readlines():
                raw_data.append(float(x))
        self.vm = neo.AnalogSignal(
            raw_data, units=pq.V, sampling_period=0.1 * pq.ms)
        self.true_time_stamps = [0.0124, 0.0354, 0.0713, 0.1192, 0.1695,
                                 0.2201, 0.2711] * pq.s

    def test_peak_detection_time_stamps(self):
        # Test with default arguments
        result = peak_detection(self.vm)
        self.assertEqual(len(self.true_time_stamps), len(result))
        self.assertIsInstance(result, neo.core.SpikeTrain)

        try:
            assert_array_almost_equal(result, self.true_time_stamps)
        except AttributeError:
            self.assertTrue(np.array_equal(result, self.true_time_stamps))

    def test_peak_detection_threshold(self):
        # Test for empty SpikeTrain when threshold is too high
        result = peak_detection(self.vm, threshold=30 * pq.mV)
        self.assertEqual(len(result), 0)


class AnalogSignalSpikeExtractionTestCase(unittest.TestCase):

    def setUp(self):
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        raw_data_file_loc = os.path.join(
            curr_dir, 'spike_extraction_test_data.txt')
        raw_data = []
        with open(raw_data_file_loc, 'r') as f:
            for x in f.readlines():
                raw_data.append(float(x))
        self.vm = neo.AnalogSignal(
            raw_data, units=pq.V, sampling_period=0.1 * pq.ms)
        self.first_spike = np.array([-0.04084546, -0.03892033, -0.03664779,
                                     -0.03392689, -0.03061474, -0.02650277,
                                     -0.0212756, -0.01443531, -0.00515365,
                                     0.00803962, 0.02797951, -0.07,
                                     -0.06974495, -0.06950466, -0.06927778,
                                     -0.06906314, -0.06885969, -0.06866651,
                                     -0.06848277, -0.06830773, -0.06814071,
                                     -0.06798113, -0.06782843, -0.06768213,
                                     -0.06754178, -0.06740699, -0.06727737,
                                     -0.06715259, -0.06703235, -0.06691635])

    def test_spike_extraction_waveform(self):
        spike_train = spike_extraction(self.vm.reshape(-1),
                                       interval=(-1 * pq.ms, 2 * pq.ms))

        assert_array_almost_equal(
            spike_train.waveforms[0][0].magnitude.reshape(-1),
            self.first_spike)


class AbstractPointProcessTestCase(unittest.TestCase):
    def test_not_implemented_error(self):
        process = AbstractPointProcess()
        self.assertRaises(
            NotImplementedError, process._generate_spiketrain_as_array)


class StationaryPoissonProcessTestCase(unittest.TestCase):
    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.

        for rate in [123.0 * pq.Hz, 0.123 * pq.kHz]:
            for t_stop in [2345 * pq.ms, 2.345 * pq.s]:
                for refractory_period in (None, 3. * pq.ms):
                    np.random.seed(seed=123456)
                    spiketrain_old = homogeneous_poisson_process(
                        rate, t_stop=t_stop,
                        refractory_period=refractory_period)
                    np.random.seed(seed=123456)

                    spiketrain = StationaryPoissonProcess(
                        rate, t_stop=t_stop,
                        refractory_period=refractory_period,
                        equilibrium=False
                    ).generate_spiketrain()
                    assert_array_almost_equal(
                        spiketrain_old.magnitude,
                        spiketrain.magnitude)
                    intervals = isi(spiketrain)

                    expected_mean_isi = 1. / rate.simplified
                    self.assertAlmostEqual(
                        expected_mean_isi.magnitude,
                        intervals.mean().simplified.magnitude,
                        places=3)

                    expected_first_spike = 0 * pq.ms
                    self.assertLess(
                        spiketrain[0] - expected_first_spike,
                        7 * expected_mean_isi)

                    expected_last_spike = t_stop
                    self.assertLess(expected_last_spike -
                                    spiketrain[-1], 7 * expected_mean_isi)

                    if refractory_period is None:
                        # Kolmogorov-Smirnov test
                        D, p = kstest(
                            intervals.rescale(t_stop.units).magnitude,
                            "expon",
                            # args are (loc, scale)
                            args=(0., expected_mean_isi.rescale(
                                t_stop.units).magnitude),
                            alternative='two-sided')
                    else:
                        refractory_period = refractory_period.rescale(
                            t_stop.units).item()
                        measured_rate = 1. / expected_mean_isi.rescale(
                            t_stop.units).item()
                        effective_rate = measured_rate / (
                                1. - measured_rate * refractory_period)

                        # Kolmogorov-Smirnov test
                        D, p = kstest(
                            intervals.rescale(t_stop.units).magnitude,
                            "expon",
                            # args are (loc, scale)
                            args=(refractory_period, 1. / effective_rate),
                            alternative='two-sided')
                    self.assertGreater(p, 0.001)
                    self.assertLess(D, 0.12)

    def test_zero_refractory_period(self):
        rate = 10 * pq.Hz
        t_stop = 20 * pq.s

        np.random.seed(27)
        sp1 = StationaryPoissonProcess(
            rate, t_stop=t_stop).generate_spiketrain(as_array=True)

        np.random.seed(27)
        sp2 = StationaryPoissonProcess(
            rate, t_stop=t_stop, refractory_period=0. * pq.ms
        ).generate_spiketrain(
            as_array=True)

        assert_array_almost_equal(sp1, sp2)

    def test_t_start_and_t_stop(self):
        rate = 10 * pq.Hz
        t_start = 17 * pq.ms
        t_stop = 2 * pq.s

        sp1 = StationaryPoissonProcess(
            rate, t_start=t_start, t_stop=t_stop).generate_spiketrain()

        sp2 = StationaryPoissonProcess(
            rate, t_start=t_start, t_stop=t_stop, refractory_period=3 * pq.ms
        ).generate_spiketrain()

        for spiketrain in (sp1, sp2):
            self.assertEqual(spiketrain.t_start, t_start)
            self.assertEqual(spiketrain.t_stop, t_stop)

    def test_zero_rate(self):
        for refractory_period in (None, 3 * pq.ms):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # RuntimeWarning: divide by zero encountered in true_divide
                # mean_interval = 1 / rate.magnitude, when rate == 0 Hz.
                spiketrain = StationaryPoissonProcess(
                    rate=0 * pq.Hz, t_stop=10 * pq.s,
                    refractory_period=refractory_period).generate_spiketrain()
                self.assertEqual(spiketrain.size, 0)

    def test_nondecrease_spike_times(self):
        for refractory_period in (None, 3 * pq.ms):
            np.random.seed(27)

            spiketrain = StationaryPoissonProcess(
                rate=10 * pq.Hz, t_stop=1000 * pq.s,
                refractory_period=refractory_period).generate_spiketrain()
            diffs = np.diff(spiketrain.times)
            self.assertTrue((diffs >= 0).all())

    def test_compare_with_as_array(self):
        rate = 10 * pq.Hz
        t_stop = 10 * pq.s
        for refractory_period in (None, 3 * pq.ms):
            process = StationaryPoissonProcess(
                rate=rate, t_stop=t_stop, refractory_period=refractory_period)
            np.random.seed(27)
            spiketrain = process.generate_spiketrain()
            self.assertIsInstance(spiketrain, neo.SpikeTrain)
            np.random.seed(27)
            spiketrain_array = process.generate_spiketrain().as_array()
            # don't check with isinstance: Quantity is a subclass of np.ndarray
            self.assertTrue(isinstance(spiketrain_array, np.ndarray))
            assert_array_almost_equal(spiketrain.times.magnitude,
                                      spiketrain_array)

    def test_effective_rate_refractory_period(self):
        np.random.seed(27)
        rate_expected = 10 * pq.Hz
        refractory_period = 90 * pq.ms  # 10 ms of effective ISI
        spiketrain = StationaryPoissonProcess(
            rate_expected, t_stop=1000 * pq.s,
            refractory_period=refractory_period
        ).generate_spiketrain()
        rate_obtained = len(spiketrain) / spiketrain.t_stop
        rate_obtained = rate_obtained.simplified
        self.assertAlmostEqual(rate_expected.simplified,
                               rate_obtained.simplified, places=1)
        intervals = isi(spiketrain)
        isi_mean_expected = 1. / rate_expected
        self.assertAlmostEqual(isi_mean_expected.simplified,
                               intervals.mean().simplified, places=3)

    def test_invalid(self):
        rate = 10 * pq.Hz
        for refractory_period in (None, 3 * pq.ms):
            # t_stop < t_start

            hpp = StationaryPoissonProcess
            self.assertRaises(
                ValueError, hpp, rate=rate, t_start=5 * pq.ms,
                t_stop=1 * pq.ms, refractory_period=refractory_period)
            # no units provided for rate, t_stop
            self.assertRaises(ValueError, hpp, rate=10,
                              refractory_period=refractory_period)
            self.assertRaises(ValueError, hpp, rate=rate, t_stop=5,
                              refractory_period=refractory_period)
            # no units provided for refractory_period
            self.assertRaises(ValueError, hpp, rate=rate, refractory_period=2)
        self.assertRaises(ValueError, StationaryPoissonProcess,
                          rate, refractory_period=1. * pq.s)


class StationaryGammaProcessTestCase(unittest.TestCase):

    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.
        a = 3.0
        for b in (67.0 * pq.Hz, 0.067 * pq.kHz):
            for t_stop in (2345 * pq.ms, 2.345 * pq.s):
                np.random.seed(seed=12345)
                spiketrain_old = homogeneous_gamma_process(
                    a, b, t_stop=t_stop)
                np.random.seed(seed=12345)
                spiketrain = StationaryGammaProcess(
                    rate=b / a, shape_factor=a, t_stop=t_stop,
                    equilibrium=False
                ).generate_spiketrain()
                assert_allclose(spiketrain_old.magnitude, spiketrain.magnitude)

                intervals = isi(spiketrain)

                expected_spike_count = int((b / a * t_stop).simplified)
                # should fail about 1 time in 1000
                self.assertLess(
                    pdiff(expected_spike_count, spiketrain.size), 0.25)

                expected_mean_isi = (a / b).rescale(pq.ms)
                self.assertLess(
                    pdiff(expected_mean_isi, intervals.mean()), 0.3)

                expected_first_spike = 0 * pq.ms
                self.assertLess(
                    spiketrain[0] - expected_first_spike,
                    4 * expected_mean_isi)

                expected_last_spike = t_stop
                self.assertLess(expected_last_spike -
                                spiketrain[-1], 4 * expected_mean_isi)

                # Kolmogorov-Smirnov test
                D, p = kstest(intervals.rescale(t_stop.units),
                              "gamma",
                              # args are (a, loc, scale)
                              args=(a, 0, (1 / b).rescale(t_stop.units)),
                              alternative='two-sided')
                self.assertGreater(p, 0.001)
                self.assertLess(D, 0.25)

    def test_compare_with_as_array(self):
        a = 3.
        b = 10 * pq.Hz
        np.random.seed(27)
        spiketrain = StationaryGammaProcess(
            rate=b / a, shape_factor=a,
            equilibrium=False).generate_spiketrain()
        self.assertIsInstance(spiketrain, neo.SpikeTrain)
        np.random.seed(27)
        spiketrain_array = StationaryGammaProcess(
            rate=b / a, shape_factor=a, equilibrium=False).generate_spiketrain(
            as_array=True)
        # don't check with isinstance: pq.Quantity is a subclass of np.ndarray
        self.assertTrue(isinstance(spiketrain_array, np.ndarray))
        assert_array_almost_equal(spiketrain.times.magnitude, spiketrain_array)


class StationaryLogNormalProcessTestCase(unittest.TestCase):

    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.
        sigma = 1.2
        for rate in (67.0 * pq.Hz, 0.067 * pq.kHz):
            for t_stop in (2345 * pq.ms, 2.345 * pq.s):
                np.random.seed(seed=123456)
                spiketrain = StationaryLogNormalProcess(
                    rate=rate, sigma=sigma, t_stop=t_stop,
                    equilibrium=False
                ).generate_spiketrain()

                intervals = isi(spiketrain)

                expected_spike_count = int((rate * t_stop).simplified)
                # should fail about 1 time in 1000
                self.assertLess(
                    pdiff(expected_spike_count, spiketrain.size), 0.25)

                expected_mean_isi = (1 / rate).rescale(pq.ms)
                self.assertLess(
                    pdiff(expected_mean_isi, intervals.mean()), 0.3)

                expected_first_spike = 0 * pq.ms
                self.assertLess(
                    spiketrain[0] - expected_first_spike,
                    4 * expected_mean_isi)

                expected_last_spike = t_stop
                self.assertLess(expected_last_spike -
                                spiketrain[-1], 4 * expected_mean_isi)

                # Kolmogorov-Smirnov test
                D, p = kstest(intervals.rescale(t_stop.units),
                              "lognorm",
                              # args are (s, loc, scale)
                              args=(sigma, 0,
                                    (1 / rate * np.exp(- sigma ** 2 / 2)
                                     ).rescale(t_stop.units)),
                              alternative='two-sided')
                self.assertGreater(p, 0.001)
                self.assertLess(D, 0.25)

    def test_compare_with_as_array(self):
        sigma = 1.2
        rate = 10 * pq.Hz
        np.random.seed(27)
        spiketrain = StationaryLogNormalProcess(
            rate=rate, sigma=sigma,
            equilibrium=False).generate_spiketrain()
        self.assertIsInstance(spiketrain, neo.SpikeTrain)
        np.random.seed(27)
        spiketrain_array = StationaryLogNormalProcess(
            rate=rate, sigma=sigma, equilibrium=False
        ).generate_spiketrain(
            as_array=True)
        # don't check with isinstance: pq.Quantity is a subclass of np.ndarray
        self.assertTrue(isinstance(spiketrain_array, np.ndarray))
        assert_array_almost_equal(spiketrain.times.magnitude, spiketrain_array)


class StationaryInverseGaussianProcessTestCase(unittest.TestCase):

    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.
        cv = 0.9
        for rate in (67.0 * pq.Hz, 0.067 * pq.kHz):
            for t_stop in (2345 * pq.ms, 2.345 * pq.s):
                np.random.seed(seed=123456)
                spiketrain = StationaryInverseGaussianProcess(
                    rate=rate, cv=cv, t_stop=t_stop, equilibrium=False
                ).generate_spiketrain()

                intervals = isi(spiketrain)

                expected_spike_count = int((rate * t_stop).simplified)
                # should fail about 1 time in 1000

                self.assertLess(
                    pdiff(expected_spike_count, spiketrain.size), 0.25)

                expected_mean_isi = (1 / rate).rescale(pq.ms)
                self.assertLess(
                    pdiff(expected_mean_isi, intervals.mean()), 0.3)

                expected_first_spike = 0 * pq.ms
                self.assertLess(
                    spiketrain[0] - expected_first_spike,
                    4 * expected_mean_isi)

                expected_last_spike = t_stop
                self.assertLess(expected_last_spike -
                                spiketrain[-1], 4 * expected_mean_isi)

                # Kolmogorov-Smirnov test
                D, p = kstest(intervals.rescale(t_stop.units),
                              "invgauss",
                              # args are (mu, loc, scale)
                              args=(cv ** 2, 0,
                                    (1 / (rate * cv ** 2)).rescale(
                                        t_stop.units)),
                              alternative='two-sided')
                self.assertGreater(p, 0.001)
                self.assertLess(D, 0.25)

    def test_compare_with_as_array(self):
        cv = 1.2
        rate = 10 * pq.Hz
        np.random.seed(27)
        spiketrain = StationaryInverseGaussianProcess(
            rate=rate, cv=cv, equilibrium=False).generate_spiketrain()
        self.assertIsInstance(spiketrain, neo.SpikeTrain)
        np.random.seed(27)
        spiketrain_array = StationaryInverseGaussianProcess(
            rate=rate, cv=cv, equilibrium=False).generate_spiketrain(
            as_array=True)
        # don't check with isinstance: pq.Quantity is a subclass of np.ndarray
        self.assertTrue(isinstance(spiketrain_array, np.ndarray))
        assert_array_almost_equal(spiketrain.times.magnitude, spiketrain_array)


class FirstSpikeCvTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(987654321)
        self.rate = 100. * pq.Hz
        self.t_stop = 10. * pq.s
        self.n_spiketrains = 10

        # can only have CV equal to 1.
        self.poisson_process = StationaryPoissonProcess(
            rate=self.rate,
            t_stop=self.t_stop)

        # choose all further processes to have CV of 1/2
        # CV = 1 - rate * refractory_period
        self.poisson_refractory_period_ordinary = StationaryPoissonProcess(
            rate=self.rate,
            refractory_period=0.5 / self.rate,
            t_stop=self.t_stop,
            equilibrium=False)

        self.poisson_refractory_period_equilibrium = \
            StationaryPoissonProcess(
                rate=self.rate,
                refractory_period=0.5 / self.rate,
                t_stop=self.t_stop,
                equilibrium=True)

        # CV = 1 / sqrt(shape_factor)
        self.gamma_process_ordinary = StationaryGammaProcess(
            rate=self.rate,
            shape_factor=4,
            t_stop=self.t_stop,
            equilibrium=False)

        self.gamma_process_equilibrium = StationaryGammaProcess(
            rate=self.rate,
            shape_factor=4,
            t_stop=self.t_stop,
            equilibrium=True)

        # CV = sqrt(exp(sigma**2) - 1)
        self.log_normal_process_ordinary = StationaryLogNormalProcess(
            rate=self.rate,
            sigma=np.sqrt(np.log(5. / 4.)),
            t_stop=self.t_stop,
            equilibrium=False)

        self.log_normal_process_equilibrium = StationaryLogNormalProcess(
            rate=self.rate,
            sigma=np.sqrt(np.log(5. / 4.)),
            t_stop=self.t_stop,
            equilibrium=True)

        self.inverse_gaussian_process_ordinary = \
            StationaryInverseGaussianProcess(
                rate=self.rate,
                cv=1 / 2,
                t_stop=self.t_stop,
                equilibrium=False)

        self.inverse_gaussian_process_equilibrium = \
            StationaryInverseGaussianProcess(
                rate=self.rate,
                cv=1 / 2,
                t_stop=self.t_stop,
                equilibrium=True)

    def test_cv(self):
        processes = (self.poisson_process,
                     self.poisson_refractory_period_ordinary,
                     self.gamma_process_ordinary,
                     self.log_normal_process_ordinary,
                     self.inverse_gaussian_process_ordinary)
        for process in processes:
            if process is self.poisson_process:
                self.assertAlmostEqual(1., process.expected_cv)

                # test the general expected-cv function
                self.assertAlmostEqual(
                    1., super(type(process), process).expected_cv)
            else:
                self.assertAlmostEqual(0.5, process.expected_cv)
                # test the general expected-cv function
                self.assertAlmostEqual(
                    0.5, super(type(process), process).expected_cv)
            spiketrains = process.generate_n_spiketrains(
                n_spiketrains=self.n_spiketrains,
                as_array=True)

            cvs = [variation(np.diff(spiketrain))
                   for spiketrain in spiketrains]
            mean_cv = np.mean(cvs)

            assert_allclose(
                process.expected_cv, mean_cv, atol=0.01)

    def test_first_spike(self):
        ordinary_processes = (self.poisson_refractory_period_ordinary,
                              self.gamma_process_ordinary,
                              self.log_normal_process_ordinary,
                              self.inverse_gaussian_process_ordinary)
        equilibrium_processes = (self.poisson_refractory_period_equilibrium,
                                 self.gamma_process_equilibrium,
                                 self.log_normal_process_equilibrium,
                                 self.inverse_gaussian_process_equilibrium)

        for ordinary_process, equilibrium_process in zip(
                ordinary_processes, equilibrium_processes):
            ordinary_spiketrains = ordinary_process.generate_n_spiketrains(
                self.n_spiketrains)
            equilbrium_spiketrains = \
                equilibrium_process.generate_n_spiketrains(
                    self.n_spiketrains)
            first_spikes_ordinary = [spiketrain[0].item()
                                     for spiketrain in ordinary_spiketrains]
            first_spikes_equilibrium = \
                [spiketrain[0].item()
                 for spiketrain in equilbrium_spiketrains]
            mean_first_spike_ordinary = np.mean(first_spikes_ordinary)
            mean_first_spike_equilibrium = np.mean(first_spikes_equilibrium)

            # for regular spike trains (CV=0.5 here) the first spike
            # in equilibrium is on average than in the ordinary case
            self.assertLess(mean_first_spike_equilibrium,
                            mean_first_spike_ordinary)


class NonStationaryPoissonProcessTestCase(unittest.TestCase):
    def setUp(self):
        rate_list = [[20]] * 1000 + [[200]] * 1000
        self.rate_profile = neo.AnalogSignal(
            rate_list * pq.Hz, sampling_period=0.001 * pq.s)
        rate_0 = [[0]] * 1000
        self.rate_profile_0 = neo.AnalogSignal(
            rate_0 * pq.Hz, sampling_period=0.001 * pq.s)
        rate_negative = [[-1]] * 1000
        self.rate_profile_negative = neo.AnalogSignal(
            rate_negative * pq.Hz, sampling_period=0.001 * pq.s)

    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.
        for rate in (self.rate_profile, self.rate_profile.rescale(pq.kHz)):
            for refractory_period in (3 * pq.ms, None):
                np.random.seed(seed=12345)
                spiketrain_old = inhomogeneous_poisson_process(
                    rate, refractory_period=refractory_period)
                np.random.seed(seed=12345)

                process = NonStationaryPoissonProcess
                spiketrain = process(rate, refractory_period=refractory_period
                                     ).generate_spiketrain()

                assert_allclose(
                    spiketrain_old.magnitude, spiketrain.magnitude)

                intervals = isi(spiketrain)

                # Computing expected statistics and percentiles
                expected_spike_count = (np.sum(
                    rate) * rate.sampling_period).simplified
                percentile_count = poisson.ppf(.999, expected_spike_count)
                expected_min_isi = (1 / np.min(rate))
                expected_max_isi = (1 / np.max(rate))
                percentile_min_isi = expon.ppf(.999, expected_min_isi)
                percentile_max_isi = expon.ppf(.999, expected_max_isi)

                # Check that minimal ISI is greater than the refractory_period
                if refractory_period is not None:
                    self.assertGreater(np.min(intervals), refractory_period)

                # Testing (each should fail 1 every 1000 times)
                self.assertLess(spiketrain.size, percentile_count)
                self.assertLess(np.min(intervals), percentile_min_isi)
                self.assertLess(np.max(intervals), percentile_max_isi)

                # Testing t_start t_stop
                self.assertEqual(rate.t_stop, spiketrain.t_stop)
                self.assertEqual(rate.t_start, spiketrain.t_start)

        # Testing type
        spiketrain_as_array = NonStationaryPoissonProcess(
            rate).generate_spiketrain(as_array=True)
        self.assertTrue(isinstance(spiketrain_as_array, np.ndarray))
        self.assertTrue(isinstance(spiketrain, neo.SpikeTrain))

        # Testing type for refractory period
        refractory_period = 3 * pq.ms
        spiketrain = NonStationaryPoissonProcess(
            rate, refractory_period=refractory_period).generate_spiketrain()
        spiketrain_as_array = NonStationaryPoissonProcess(
            rate, refractory_period=refractory_period).generate_spiketrain(
            as_array=True)
        self.assertTrue(isinstance(spiketrain_as_array, np.ndarray))
        self.assertTrue(isinstance(spiketrain, neo.SpikeTrain))

        # Check that to high refractory period raises error
        self.assertRaises(
            ValueError, NonStationaryPoissonProcess,
            self.rate_profile,
            refractory_period=1000 * pq.ms)

    def test_effective_rate_refractory_period(self):
        np.random.seed(27)
        rate_expected = 10 * pq.Hz
        refractory_period = 90 * pq.ms  # 10 ms of effective ISI
        rates = neo.AnalogSignal(np.repeat(rate_expected, 1000), units=pq.Hz,
                                 t_start=0 * pq.ms, sampling_rate=1 * pq.Hz)
        spiketrain = NonStationaryPoissonProcess(
            rates, refractory_period=refractory_period).generate_spiketrain()
        rate_obtained = len(spiketrain) / spiketrain.t_stop
        self.assertAlmostEqual(
            rate_expected.simplified.item(),
            rate_obtained.simplified.item(), places=1)
        intervals_inhomo = isi(spiketrain)
        isi_mean_expected = 1. / rate_expected
        self.assertAlmostEqual(isi_mean_expected.simplified,
                               intervals_inhomo.mean().simplified,
                               places=3)

    def test_zero_rate(self):
        for refractory_period in (3 * pq.ms, None):
            process = NonStationaryPoissonProcess
            spiketrain = process(
                self.rate_profile_0, refractory_period=refractory_period
            ).generate_spiketrain()
            self.assertEqual(spiketrain.size, 0)
        self.assertRaises(
            ValueError, NonStationaryPoissonProcess,
            self.rate_profile, refractory_period=5)

    def test_negative_rates(self):
        for refractory_period in (3 * pq.ms, None):
            process = NonStationaryPoissonProcess
            self.assertRaises(
                ValueError, process,
                self.rate_profile_negative,
                refractory_period=refractory_period)


class NonStationaryGammaTestCase(unittest.TestCase):
    def setUp(self):
        rate_list = [[20]] * 1000 + [[200]] * 1000
        self.rate_profile = neo.AnalogSignal(
            rate_list * pq.Hz, sampling_period=0.001 * pq.s)
        rate_0 = [[0]] * 1000
        self.rate_profile_0 = neo.AnalogSignal(
            rate_0 * pq.Hz, sampling_period=0.001 * pq.s)
        rate_negative = [[-1]] * 1000
        self.rate_profile_negative = neo.AnalogSignal(
            rate_negative * pq.Hz, sampling_period=0.001 * pq.s)

    def test_statistics(self):
        # This is a statistical test that has a non-zero chance of failure
        # during normal operation. Thus, we set the random seed to a value that
        # creates a realization passing the test.
        shape_factor = 2.5

        for rate in [self.rate_profile, self.rate_profile.rescale(pq.kHz)]:
            np.random.seed(seed=12345)
            spiketrain_old = inhomogeneous_gamma_process(
                rate, shape_factor=shape_factor)
            np.random.seed(seed=12345)
            spiketrain = NonStationaryGammaProcess(
                rate, shape_factor=shape_factor).generate_spiketrain()
            assert_allclose(spiketrain_old.magnitude, spiketrain.magnitude)

            intervals = isi(spiketrain)

            # Computing expected statistics and percentiles
            expected_spike_count = (np.sum(
                rate) * rate.sampling_period).simplified
            percentile_count = poisson.ppf(.999, expected_spike_count)
            expected_min_isi = (1 / np.min(rate))
            expected_max_isi = (1 / np.max(rate))
            percentile_min_isi = expon.ppf(.999, expected_min_isi)
            percentile_max_isi = expon.ppf(.999, expected_max_isi)

            # Testing (each should fail 1 every 1000 times)
            self.assertLess(spiketrain.size, percentile_count)
            self.assertLess(np.min(intervals), percentile_min_isi)
            self.assertLess(np.max(intervals), percentile_max_isi)

            # Testing t_start t_stop
            self.assertEqual(rate.t_stop, spiketrain.t_stop)
            self.assertEqual(rate.t_start, spiketrain.t_start)

        # Testing type
        spiketrain_as_array = NonStationaryGammaProcess(
            rate, shape_factor=shape_factor).generate_spiketrain(
            as_array=True)
        self.assertTrue(isinstance(spiketrain_as_array, np.ndarray))
        self.assertTrue(isinstance(spiketrain, neo.SpikeTrain))

        # check error if rate has wrong format
        self.assertRaises(
            ValueError, NonStationaryGammaProcess,
            rate_signal=[0.1, 2.],
            shape_factor=shape_factor)

        # check error if negative values in rate
        self.assertRaises(
            ValueError, NonStationaryGammaProcess,
            rate_signal=neo.AnalogSignal(
                [-0.1, 10.] * pq.Hz, sampling_period=0.001 * pq.s),
            shape_factor=shape_factor)

        # check error if rate is empty
        self.assertRaises(
            ValueError, NonStationaryGammaProcess,
            rate_signal=neo.AnalogSignal(
                [] * pq.Hz, sampling_period=0.001 * pq.s),
            shape_factor=shape_factor)

    def test_recovered_firing_rate_profile(self):
        np.random.seed(54)
        t_start = 0 * pq.s
        t_stop = 2 * np.round(np.pi, decimals=3) * pq.s  # 2 full periods
        sampling_period = 0.001 * pq.s

        # an arbitrary rate profile
        profile = 0.5 * (1 + np.sin(np.arange(t_start.item(), t_stop.item(),
                                              sampling_period.item())))

        n_trials = 100
        rtol = 0.1  # 10% of deviation allowed
        kernel = kernels.RectangularKernel(sigma=0.25 * pq.s)
        for rate in (10 * pq.Hz, 50 * pq.Hz):
            rate_profile = neo.AnalogSignal(rate * profile,
                                            sampling_period=sampling_period)
            # the recovered firing rate profile should not depend on the
            # shape factor; here we test float and integer values of the shape
            # factor: the method supports float values that is not trivial
            # for inhomogeneous gamma process generation
            for shape_factor in (2.5, 10.):
                spiketrains = NonStationaryGammaProcess(
                    rate_profile, shape_factor=shape_factor
                ).generate_n_spiketrains(n_trials)
                rate_recovered = instantaneous_rate(
                    spiketrains,
                    sampling_period=sampling_period,
                    kernel=kernel,
                    t_start=t_start,
                    t_stop=t_stop, trim=True).sum(axis=1) / n_trials

                rate_recovered = rate_recovered.flatten().magnitude
                trim = (rate_profile.shape[0] - rate_recovered.shape[0]) // 2
                rate_profile_valid = rate_profile.magnitude.squeeze()
                rate_profile_valid = rate_profile_valid[trim: -trim]
                assert_allclose(rate_recovered, rate_profile_valid,
                                rtol=0, atol=rtol * rate.item())


class NPoissonTestCase(unittest.TestCase):

    def setUp(self):
        self.n = 4
        self.rate = 10 * pq.Hz
        self.rates = np.arange(1, self.n + 1) * pq.Hz
        self.t_stop = 10000 * pq.ms

    def test_poisson(self):
        # Check the output types for input rate + n number of neurons
        pp = _n_poisson(
            rate=self.rate,
            t_stop=self.t_stop,
            n_spiketrains=self.n)
        self.assertIsInstance(pp, list)
        self.assertIsInstance(pp[0], neo.core.spiketrain.SpikeTrain)
        self.assertEqual(pp[0].simplified.units, 1000 * pq.ms)
        self.assertEqual(len(pp), self.n)

        # Check the output types for input list of rates
        pp = _n_poisson(rate=self.rates, t_stop=self.t_stop)
        self.assertIsInstance(pp, list)
        self.assertIsInstance(pp[0], neo.core.spiketrain.SpikeTrain)
        self.assertEqual(pp[0].simplified.units, 1000 * pq.ms)
        self.assertEqual(len(pp), self.n)

    def test_poisson_error(self):
        # Dimensionless rate
        self.assertRaises(
            ValueError, _n_poisson, rate=5, t_stop=self.t_stop)
        # Negative rate
        self.assertRaises(
            ValueError, _n_poisson, rate=-5 * pq.Hz, t_stop=self.t_stop)
        # Negative value when rate is a list
        self.assertRaises(
            ValueError, _n_poisson, rate=[-5, 3] * pq.Hz,
            t_stop=self.t_stop)
        # Negative n
        self.assertRaises(
            ValueError, _n_poisson, rate=self.rate, t_stop=self.t_stop,
            n_spiketrains=-1)
        # t_start>t_stop
        self.assertRaises(
            ValueError, _n_poisson, rate=self.rate, t_start=4 * pq.ms,
            t_stop=3 * pq.ms, n_spiketrains=3)


class SingleInteractionProcessTestCase(unittest.TestCase):
    def setUp(self):
        self.n = 4
        self.rate = 10 * pq.Hz
        self.rates = np.arange(1, self.n + 1) * pq.Hz
        self.t_stop = 10000 * pq.ms
        self.rate_c = 1 * pq.Hz

    def format_check(self, sip, coinc):
        self.assertEqual(type(sip), list)
        self.assertEqual(type(sip[0]), neo.core.spiketrain.SpikeTrain)
        self.assertEqual(type(coinc[0]), neo.core.spiketrain.SpikeTrain)
        self.assertEqual(sip[0].simplified.units, 1000 * pq.ms)
        self.assertEqual(coinc[0].simplified.units, 1000 * pq.ms)

        # Check the output length
        self.assertEqual(len(sip), self.n)

    def test_sip(self):
        # Generate an example SIP mode
        sip, coinc = single_interaction_process(
            n_spiketrains=self.n, t_stop=self.t_stop, rate=self.rate,
            coincidence_rate=self.rate_c, return_coincidences=True)

        # Check the output types
        self.format_check(sip, coinc)
        self.assertEqual(
            len(coinc[0]), (self.rate_c * self.t_stop).simplified.magnitude)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Generate an example SIP mode giving a list of rates as imput
            sip, coinc = single_interaction_process(
                t_stop=self.t_stop, rate=self.rates,
                coincidence_rate=self.rate_c, return_coincidences=True)

        # Check the output types
        self.format_check(sip, coinc)
        self.assertEqual(
            len(coinc[0]),
            (self.rate_c * self.t_stop).rescale(pq.dimensionless))

        # Generate an example SIP mode stochastic number of coincidences
        sip = single_interaction_process(
            n_spiketrains=self.n,
            t_stop=self.t_stop,
            rate=self.rate,
            coincidence_rate=self.rate_c,
            coincidences='stochastic',
            return_coincidences=False)

        # Check the output types
        self.assertEqual(type(sip), list)
        self.assertEqual(type(sip[0]), neo.core.spiketrain.SpikeTrain)
        self.assertEqual(sip[0].simplified.units, 1000 * pq.ms)

    def test_sip_error(self):
        # Negative rate
        self.assertRaises(
            ValueError, single_interaction_process, n_spiketrains=self.n,
            rate=-5 * pq.Hz,
            coincidence_rate=self.rate_c, t_stop=self.t_stop)
        # Negative coincidence rate
        self.assertRaises(
            ValueError, single_interaction_process, n_spiketrains=self.n,
            rate=self.rate, coincidence_rate=-3 * pq.Hz, t_stop=self.t_stop)
        # Negative value when rate is a list
        self.assertRaises(
            ValueError, single_interaction_process, n_spiketrains=self.n,
            rate=[-5, 3, 4, 2] * pq.Hz, coincidence_rate=self.rate_c,
            t_stop=self.t_stop)
        # Negative n
        self.assertRaises(
            ValueError, single_interaction_process, n_spiketrains=-1,
            rate=self.rate, coincidence_rate=self.rate_c, t_stop=self.t_stop)
        # Rate_c < rate
        self.assertRaises(
            ValueError,
            single_interaction_process,
            n_spiketrains=self.n,
            rate=self.rate,
            coincidence_rate=self.rate + 1 * pq.Hz,
            t_stop=self.t_stop)


class CppTestCase(unittest.TestCase):

    def format_check(self, cpp, amplitude_distribution, t_start, t_stop):
        self.assertEqual(
            [type(train) for train in cpp],
            [neo.SpikeTrain] * len(cpp))
        self.assertEqual(cpp[0].simplified.units, 1000 * pq.ms)
        self.assertEqual(type(cpp), list)
        # testing quantities format of the output
        self.assertEqual(
            [train.simplified.units for train in cpp],
            [1000 * pq.ms] * len(cpp))
        # testing output t_start t_stop
        for spiketrain in cpp:
            self.assertEqual(spiketrain.t_stop, t_stop)
            self.assertEqual(spiketrain.t_start, t_start)
        self.assertEqual(len(cpp), len(amplitude_distribution) - 1)

    def test_cpp_hom(self):
        # testing output with generic inputs
        amplitude_distribution = np.array([0, .9, .1])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 3 * pq.Hz
        cpp_hom = cpp(rate, amplitude_distribution,
                      t_stop, t_start=t_start)
        # testing the output formats
        self.format_check(cpp_hom, amplitude_distribution, t_start, t_stop)

        # testing the units
        t_stop = 10000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 3 * pq.Hz
        cpp_unit = cpp(rate, amplitude_distribution,
                       t_stop, t_start=t_start)

        self.assertEqual(cpp_unit[0].units, t_stop.units)
        self.assertEqual(cpp_unit[0].t_stop.units, t_stop.units)
        self.assertEqual(cpp_unit[0].t_start.units, t_stop.units)

        # testing output without copy of spikes
        amplitude_distribution = np.array([1])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 3 * pq.Hz
        cpp_hom_empty = cpp(
            rate, amplitude_distribution, t_stop, t_start=t_start)

        self.assertEqual(
            [len(train) for train in cpp_hom_empty], [0] * len(cpp_hom_empty))

        # testing output with rate equal to 0
        amplitude_distribution = np.array([0, .9, .1])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 0 * pq.Hz
        cpp_hom_empty_r = cpp(
            rate, amplitude_distribution, t_stop, t_start=t_start)
        self.assertEqual(
            [len(train) for train in cpp_hom_empty_r], [0] * len(
                cpp_hom_empty_r))

        # testing output with same spike trains in output
        amplitude_distribution = np.array([0., 0., 1.])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 3 * pq.Hz
        cpp_hom_eq = cpp(
            rate, amplitude_distribution, t_stop, t_start=t_start)

        self.assertTrue(
            np.allclose(cpp_hom_eq[0].magnitude, cpp_hom_eq[1].magnitude))

    def test_cpp_hom_errors(self):
        # testing raises of ValueError (wrong inputs)
        # testing empty amplitude
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[],
            t_stop=10 * 1000 * pq.ms,
            rate=3 * pq.Hz)

        # testing sum of amplitude>1
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[1, 1, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=3 * pq.Hz)
        # testing negative value in the amplitude
        self.assertRaises(
            ValueError, cpp, amplitude_distribution=[-1, 1, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=3 * pq.Hz)
        # test negative rate
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Catches RuntimeWarning: invalid value encountered in sqrt
            # number = np.ceil(n + 3 * np.sqrt(n)), when `n` == -3 Hz.
            self.assertRaises(
                ValueError, cpp, amplitude_distribution=[0, 1, 0],
                t_stop=10 * 1000 * pq.ms,
                rate=-3 * pq.Hz)
        # test wrong unit for rate
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=3 * 1000 * pq.ms)

        # testing raises of AttributeError (missing input units)
        # Testing missing unit to t_stop
        self.assertRaises(
            ValueError, cpp, amplitude_distribution=[0, 1, 0], t_stop=10,
            rate=3 * pq.Hz)
        # Testing missing unit to t_start
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=3 * pq.Hz,
            t_start=3)
        # testing rate missing unit
        self.assertRaises(
            AttributeError, cpp, amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=3)

    def test_cpp_het(self):
        # testing output with generic inputs
        amplitude_distribution = np.array([0, .9, .1])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = [3, 4] * pq.Hz
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Catch RuntimeWarning: divide by zero encountered in true_divide
            # mean_interval = 1 / rate.magnitude, when rate == 0 Hz.
            cpp_het = cpp(rate, amplitude_distribution,
                          t_stop, t_start=t_start)
            # testing the output formats
            self.format_check(cpp_het, amplitude_distribution, t_start, t_stop)
            self.assertEqual(len(cpp_het), len(rate))

            # testing the units
            t_stop = 10000 * pq.ms
            t_start = 5 * 1000 * pq.ms
            rate = [3, 4] * pq.Hz
            cpp_unit = cpp(
                rate, amplitude_distribution, t_stop, t_start=t_start)

            self.assertEqual(cpp_unit[0].units, t_stop.units)
            self.assertEqual(cpp_unit[0].t_stop.units, t_stop.units)
            self.assertEqual(cpp_unit[0].t_start.units, t_stop.units)
            # testing without copying any spikes
            amplitude_distribution = np.array([1, 0, 0])
            t_stop = 10 * 1000 * pq.ms
            t_start = 5 * 1000 * pq.ms
            rate = [3, 4] * pq.Hz
            cpp_het_empty = cpp(
                rate, amplitude_distribution, t_stop, t_start=t_start)

            self.assertEqual(len(cpp_het_empty[0]), 0)

            # testing output with rate equal to 0
            amplitude_distribution = np.array([0, .9, .1])
            t_stop = 10 * 1000 * pq.ms
            t_start = 5 * 1000 * pq.ms
            rate = [0, 0] * pq.Hz
            cpp_het_empty_r = cpp(
                rate, amplitude_distribution, t_stop, t_start=t_start)
            self.assertEqual(
                [len(train) for train in cpp_het_empty_r], [0] * len(
                    cpp_het_empty_r))

            # testing completely synchronous spike trains
            amplitude_distribution = np.array([0, 0, 1])
            t_stop = 10 * 1000 * pq.ms
            t_start = 5 * 1000 * pq.ms
            rate = [3, 3] * pq.Hz
            cpp_het_eq = cpp(
                rate, amplitude_distribution, t_stop, t_start=t_start)

            self.assertTrue(np.allclose(
                cpp_het_eq[0].magnitude, cpp_het_eq[1].magnitude))

    def test_cpp_het_err(self):
        # testing raises of ValueError (wrong inputs)
        # testing empty amplitude
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz)
        # testing sum amplitude>1
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[1, 1, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz)
        # testing amplitude negative value
        self.assertRaises(
            ValueError, cpp, amplitude_distribution=[-1, 1, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz)
        # testing negative rate
        self.assertRaises(ValueError, cpp, amplitude_distribution=[
            0, 1, 0], t_stop=10 * 1000 * pq.ms, rate=[-3, 4] * pq.Hz)
        # testing empty rate
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=[] * pq.Hz)
        # testing empty amplitude
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz)
        # testing different len(A)-1 and len(rate)
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz)
        # testing rate with different unit from Hz
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * 1000 * pq.ms)
        # Testing analytical constrain between amplitude and rate
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 0, 1],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz,
            t_start=3)

        # testing raises of AttributeError (missing input units)
        # Testing missing unit to t_stop
        self.assertRaises(
            ValueError, cpp, amplitude_distribution=[0, 1, 0], t_stop=10,
            rate=[3, 4] * pq.Hz)
        # Testing missing unit to t_start
        self.assertRaises(
            ValueError,
            cpp,
            amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4] * pq.Hz,
            t_start=3)
        # Testing missing unit to rate
        self.assertRaises(
            AttributeError, cpp, amplitude_distribution=[0, 1, 0],
            t_stop=10 * 1000 * pq.ms,
            rate=[3, 4])

    def test_cpp_jttered(self):
        # testing output with generic inputs
        amplitude_distribution = np.array([0, .9, .1])
        t_stop = 10 * 1000 * pq.ms
        t_start = 5 * 1000 * pq.ms
        rate = 3 * pq.Hz
        cpp_shift = cpp(
            rate,
            amplitude_distribution,
            t_stop,
            t_start=t_start,
            shift=3 * pq.ms)
        # testing the output formats
        self.format_check(cpp_shift, amplitude_distribution, t_start, t_stop)


if __name__ == '__main__':
    unittest.main()
